#!/usr/bin/env python3
'''\
Test the Mesh object fitting, saving, and loading
'''

######## Imports ########

import numpy as np
from gwalk.model.parameter import Parameter

######## Settings ########

limits = np.asarray([
                     [0., 1.],
                    ])

mu = np.asarray([0.5])

n_rvs = int(1e6)

seed = 0
rs = np.random.RandomState(seed)

filename = "test_mesh.hdf5"

######## Generate Samples ########

P = Parameter("myparam", mu, limits, random_state=rs)



#!/home/xevra/.local/bin/python3
'''\
Generate the likelihood function of each event
'''

######## Imports ########
import numpy as np
#from scipy.stats import gaussian_kde
from os.path import join, isdir, isfile
from .model import Parameter, MultivariateNormal
from .data.database import Database
#from .plots.plot_sample_contour_simple import corner_cross_sample_normal
#from .plots.plot_sample_contour import corner_cross_sample_normal
from .catalog.coordinates import coord_tags, coord_labels
from .catalog.coordinates import m1m2_from_mc_eta
from .catalog.coordinates import mc_eta_from_m1m2
from .utils.multivariate_normal import params_of_mu_cov
from .utils.multivariate_normal import mu_of_params
from .utils.multivariate_normal import std_of_params
from .utils.multivariate_normal import cor_of_params
from .utils.multivariate_normal import cov_of_params
from .utils.multivariate_normal import cov_of_std_cor

######## Functions ########


######## NAL Object ########

class NAL(object):
    '''\
    Normal Approximate Likelihood object
    
    Inputs:
        hyperparameters: hyperparameters for the fit
        seed: random number generator seed for some things
        hist_bins: The number of bins to use in histograms

    methods:
    '''
    def __init__(
                 self,
                 fname_catalog,
                 fname_mesh,
                 fname_nal,
                 event,
                 group,
                 coord_tag,
                 evaluation_res=100,
                 seed=20200610,
                 verbose=False,
                 **mesh_kwargs
                ):
        '''\
        Initialize NAL object
        '''
        from .catalog import Catalog
        #print("Initializing NAL object for %s %s %s"%(event, group, coord_tag))
        #### Initialize random state ####
        self.rs = np.random.RandomState(seed=seed)

        #### Save attributes ####
        self.seed = seed
        self.event = event
        self.group = group
        self.coord_tag = coord_tag
        self.fname_catalog = fname_catalog
        self.fname_mesh = fname_mesh
        self.fname_nal = fname_nal
        self.verbose = verbose

        # Check to make sure file exists
        if not isfile(fname_catalog):
            raise RuntimeError("Cannot find %s"%(fname_catalog))

        # Handle coordinates
        if not coord_tag in coord_tags:
            raise RuntimeError("Unknown coord tag: %s"%coord_tag)
        # Generate additional attributes
        self.coords = coord_tags[coord_tag][:-1]
        self.prior = coord_tags[coord_tag][-1]
        self.ndim = len(self.coords)
        self.coord_labels = []
        for i in range(self.ndim):
            self.coord_labels.append(coord_labels[self.coords[i]])

        # Initialize a catalog object
        self.cata = Catalog(self.fname_catalog)

        # Check waveform/group
        assert self.cata.group_status(self.event,self.group,self.coords)

        ## Initialize release database for runs ##
        # Identify the file for the release fits
        self.db = Database(self.fname_nal)
        # Ensure the event group exists
        if not self.db.exists(event):
            self.db.create_group(event)

        # Save mesh kwargs
        self.mesh_kwargs = mesh_kwargs
        # Identify mesh label
        self.mesh_label = "%s/%s:%s"%(self.event,self.coord_tag,self.group)
        # Identify mesh resolution
        self.evaluation_res = evaluation_res
        # Initialize the mesh
        self.initialize_mesh()
        # Initialize normal object
        self.initialize_norm()

    def initialize_mesh(self):
        '''\
        Check if mesh exists, and if not, create one
        '''
        from .density import Mesh

        # Initialize new mesh flag
        new_mesh = False
        # Check for existing mesh
        if Mesh.exists(self.fname_mesh, self.mesh_label):
            # Load existing mesh
            loaded_mesh = Mesh.load(self.fname_mesh, self.mesh_label)
            # Loop through each kwarg
            for item in self.mesh_kwargs:
                # If the kwarg has changed, do not keep
                if item in loaded_mesh.attrs:
                    if not (self.mesh_kwargs[item] == loaded_mesh.attrs[item]):
                        new_mesh = True
                elif item == "limits":
                    limits = np.asarray(self.mesh_kwargs["limits"])
                    if not (limits.size == loaded_mesh.limits.size):
                        new_mesh = True
                    elif not all(limits = loaded_mesh.limits):
                        new_mesh = True

            # Keep the loaded mesh if appropriate
            if not new_mesh:
                self.mesh = loaded_mesh
                self.limits = loaded_mesh.limits
        else:
            new_mesh = True
        # Initialize new mesh
        if new_mesh:
            # Load appropriate catalog samples
            samples, inv_prior = self.load_catalog_samples()
            # Initialize mesh
            self.mesh = Mesh.fit(
                                 samples,
                                 self.ndim,
                                 weights=inv_prior,
                                 verbose=self.verbose,
                                 **self.mesh_kwargs
                                )
            # Save the mesh
            self.mesh.save(self.fname_mesh, self.mesh_label)
            # Find limits
            self.limits = self.mesh.limits

        # Generate evaluation set
        self.mesh.generate_evaluation_set(self.evaluation_res)

    def initialize_norm(self):
        '''\
        Initialize a Multivariate Normal object
        '''
        # Load things we need 
        self.variables = self.load_variables()
        # Initialize the scale
        self.scale = self.load_scale()
        # Initialize normal object
        self.norm = MultivariateNormal(
                                       self.variables,
                                       self.scale,
                                       random_state=self.rs
                                      )

    def load_catalog_samples(self):
        '''\
        Load catalog samples for us
        '''
        # Load appropriate catalog samples
        sample_dict = self.cata.load_data(self.event,self.group,self.coords)
        samples = []
        for item in self.coords:
            samples.append(sample_dict[item])
        samples = np.asarray(samples).T
        # Load prior weights
        prior = self.cata.load_data(
                                    self.event,
                                    self.group,
                                    [self.prior],
                                   )[self.prior]
        inv_prior = prior**-1
        # Return quantitites
        return samples, inv_prior

    def load_variables(self):
        '''\
        Initialize variable objects
        '''
        # Initialize list of variables
        variables = []

        for i in range(self.ndim):
            # We use this a couple times
            coord = self.coords[i]
            # Find the peak
            x_test, y_test = self.mesh.fetch_1d_evaluations(i)
            peak = x_test[1:-1][np.argmax(y_test[1:-1])]
            # Initialize a parameter
            p = Parameter(
                          coord,
                          peak,
                          self.limits[i],
                          label=self.coord_labels[i],
                         )
            # Append parameter
            variables.append(p)

        return variables

    def load_scale(self):
        '''\
        Check the scale of our samples

        '''
        # Load appropriate catalog samples
        samples, inv_prior = self.load_catalog_samples()
        # find the covariance
        cov = np.cov(samples.T, aweights=inv_prior)
        # Find the standard deviations
        std_dev = np.sqrt(np.diag(cov))
        return std_dev

    def kl_divergence(self,X=None,**kwargs):
        '''\
        Take the kl divergence between the kde grid and the fit
        '''
        #import time
        # Load parameters
        if X is None:
            X = self.norm.read_guess()
        # Find kl divergence
        kl = self.norm.mesh_kl(self.mesh,X,**kwargs)
        #print("kl: %f"%kl)
        return kl

    #### Data Handling ####
    def save_fit(
                 self,
                 fit_method,
                 save_rule='check_kl',
                 **kwargs
                ):
        '''\
        First compare current fit to what's saved, and save better
        Inputs:
            fname_db: location for .nal.hdf5 databse to live
            group: string which points to location in database
            hyp_new: fit parameters for new gaussian
        '''
        # Generate fit label
        fit_label = "%s:%s"%(self.mesh_label, fit_method)

        # Generate address
        if not self.db.exists(fit_label):
            self.db.create_group(fit_label)

        # Load mean and cov for fit
        mean, cov = self.norm.read_physical()

        # Calculate standard deviation
        std = np.sqrt(np.diag(cov))

        # Calculate corelation params
        cor = self.norm.cor_of_params(self.norm.read_guess())[0]

        # Calculate kl divergence
        this_kl = self.kl_divergence(**kwargs)

        # Initialize save flag
        save = False
        
        # Check save rule
        if save_rule == "check_kl":

            # Check if the kl divergence exists
            if not self.db.attr_exists(fit_label, "kl"):
                if self.verbose:
                    print("No existing model found. Saving new fit")
                save = True
            else:
                # Load the old fit
                old_fit = NAL.load_fit(
                                       self.fname_nal,
                                       self.group,
                                       self.event,
                                       self.coord_tag,
                                       fit_method,
                                      )
                
                # Load the old fit parameters
                X_old = old_fit.read_guess()

                # First check if the scale is the same
                if not (self.scale.size == old_fit.scale.size):
                    save = True
                elif any(self.scale != old_fit.scale):
                    save = True
                else:
                    # Generate the old_fit kl divergence
                    old_kl = self.kl_divergence(X=X_old,**kwargs)

                
                    if this_kl < old_kl:
                        save = True
                        if self.verbose:
                            print("Existing model found. Saving improvement")
                            print("%f < %f"%(this_kl, old_kl))
                    else:
                        save = False
                        if self.verbose:
                            print("Existing model found. Fit is not better. Not saving")
                            print("%f > %f"%(this_kl, old_kl))

        elif save_rule == "never":
            save = False
        elif save_rule == "always":
            save = True
        else:
            raise RuntimeError("Unknown save rule: %s"%save_rule)

        if save:
            # Set datasets
            self.db.dset_set(join(fit_label,"scale"),    self.scale)
            self.db.dset_set(join(fit_label,"mean"),     mean)
            self.db.dset_set(join(fit_label,"cov"),      cov)
            self.db.dset_set(join(fit_label,"std"),      std)
            self.db.dset_set(join(fit_label,"cor"),      cor)
            self.db.dset_set(join(fit_label, "limits"),  self.limits)
            # Set attributes
            self.db.attr_set(fit_label, "kl", this_kl)
            self.db.attr_set(fit_label, "ndim", self.ndim)
            self.db.attr_set(fit_label, "coords", self.coords)
            self.db.attr_set(fit_label, "seed", self.seed)
            self.db.attr_set(fit_label, "event", self.event)
            self.db.attr_set(fit_label, "group", self.group)
            # Set mesh things
            self.db.attr_set(fit_label, "mesh_evaluation_res", self.evaluation_res)
            self.db.attr_set(fit_label, "mesh_min_bins", self.mesh.attrs["min_bins"])
            self.db.attr_set(fit_label, "mesh_max_bins1d", self.mesh.attrs["max_bins1d"])
            self.db.attr_set(fit_label, "mesh_max_bins2d", self.mesh.attrs["max_bins2d"])
            self.db.attr_set(fit_label, "mesh_whitenoise", self.mesh.attrs["whitenoise"])
            self.db.attr_set(fit_label, "mesh_grab_edge", self.mesh.attrs["grab_edge"])
            self.db.attr_set(fit_label, "mesh_gp_order", self.mesh.attrs["order"])
            # Set mesh attributes
            for item in kwargs:
                self.db.attr_set(fit_label, item, kwargs[item])
                
    #### Fit event ####
    def fit_simple(
                   self,
                   save_rule="always",
                   assign=True,
                   **hyperparameters
                  ):
        '''\
        Main fit function for events
        '''
        import time
        # Generate fit label
        fit_label = "%s:%s"%(self.mesh_label, "simple")

        # Inform the user we are about to fit
        if self.verbose:
            print("Fitting Normal Model - %s"%self.event)
            print("  fit label: %s"%fit_label)
            print("  coordinates: %s"%(str(self.coords)))

        # Find the start time for the fit
        t_start = time.time()

        ### Fit Gaussians ###
        # Load appropriate catalog samples
        samples, inv_prior = self.load_catalog_samples()
        # Fit the catalog samples
        self.norm.fit_simple(samples,w=inv_prior,assign=assign)

        # End time
        t_end = time.time()
        fit_time = t_end - t_start
        hyperparameters["fit_time"] = fit_time
        
        # Inform the user about the fit
        if self.verbose:
            seconds = int(fit_time % 60)
            minutes = int((fit_time / 60) % 60)
            hours = int(fit_time/3600)
            print("\nTime: %d hours, %d minutes, %d seconds"%(hours, minutes, seconds))

            # Calculate kl divergence
            this_kl = self.kl_divergence()
            print("kl: ", this_kl)
            
            # Load mean and cov for fit
            mean_cur, cov_cur = self.norm.read_physical()
            print("mean", mean_cur)
            print("cov: ")
            print(cov_cur)

            # Calculate standard deviation
            std_dev = np.sqrt(np.diag(cov_cur))
            print("std_dev: ", std_dev)

            # Calculate corelation params
            cor = self.norm.cor_of_params(self.norm.read_guess())
            print("Correlation:")
            print(cor)

        # Save simple fit
        if self.verbose and not (save_rule == "never"):
            print("Saving fit")

            self.save_fit(
                          "simple",
                          save_rule=save_rule,
                          **hyperparameters
                         )

    def fit_event(
                  self,
                  fit_method,
                  save_rule="check_kl",
                  **hyperparameters
                 ):
        '''\
        Main fit function for events
        '''
        import time
        # Generate fit label
        fit_label = "%s:%s"%(self.mesh_label, fit_method)

        # Inform the user we are about to fit
        if self.verbose:
            print("Fitting Normal Model - %s"%self.event)
            print("  fit label: %s"%fit_label)
            print("  coordinates: %s"%(str(self.coords)))

        # Find the start time for the fit
        t_start = time.time()

        ### Fit Gaussians ###
        if fit_method in ["simple"]:
            # Load appropriate catalog samples
            samples, inv_prior = self.load_catalog_samples()
            # Fit the catalog samples
            self.norm.fit_simple(samples,w=inv_prior)

        ## Genetic Algorithm ##
        elif fit_method == "genetic":
            # Make guesses
            Xg, kl_sum = self.generate_guesses(
                           kl_sensitivity=hyperparameters["kl_sensitivity"],
                          )
            # Perform fit
            self.norm.fit_random_walk(
                                      mesh=self.mesh,
                                      fit_method = fit_method,
                                      nwalk = hyperparameters["nwalk"],
                                      nstep = hyperparameters["nstep"],
                                      kl_sensitivity = hyperparameters["kl_sensitivity"],
                                      carryover = hyperparameters["carryover"],
                                      convergence=hyperparameters["convergence"],
                                      verbose=self.verbose,
                                      guess_list=Xg,
                                     )
        ## Guess ##
        elif fit_method == "select":
            # Make guesses
            Xg, kl_sum = self.generate_guesses(
                           kl_sensitivity=hyperparameters["kl_sensitivity"],
                          )
            # Assign the best guess
            self.norm.assign_guess(Xg[0])

        """

        ## Simulated Annealing ##
        elif fit_method == "annealing":
            # load catalog samples
            posterior, prior = self.grid.load_catalog_samples()
            # Generate inverse prior
            inv_prior = np.power(prior, -1.)
            # Make intelligent guesses
            guess_list, kl_sum = \
                make_guess(Norm, self.grid, nfill=hyperparameters["nwalk"])
            # Load grid information
            cata_hist_data = self.grid.load_catalog_hist()
            # Ensure ntau is not too high
            if hyperparameters["ntau"] > hyperparameters["nstep"] / hyperparameters["nwalk"]:
                hyperparameters["ntau"] = hyperparameters["nstep"] / hyperparameters["nwalk"]
            # Load limits
            limits = Norm.read_limits()
            # Do not start too zoomed out
            guess_scale = (np.asarray(guess_list).max(axis=0) - \
                           np.asarray(guess_list).min(axis=0)) / \
                          (limits[:,1] - limits[:,0])

            # Perform fit
            Norm.fit_random_walk(
                                 fit_method = fit_method,
                                 nwalk = hyperparameters["nwalk"],
                                 nstep = hyperparameters["nstep"],
                                 ntau = hyperparameters["ntau"],
                                 jump_scale = guess_scale,
                                 kl_bins = self.grid.kl_bins,
                                 kl_sensitivity = self.grid.kl_sensitivity,
                                 fit_samples = self.grid.fit_samples,
                                 hist_data = cata_hist_data,
                                 coords = self.grid.coords,
                                 guess_list = guess_list,
                                 replacement = hyperparameters["replacement"],
                                )

        ## Genetic Algorithm ##
        elif fit_method == "genetic":
            # load catalog samples
            posterior, prior = self.grid.load_catalog_samples()
            # Generate inverse prior
            inv_prior = np.power(prior, -1.)
            # Make intelligent guesses
            guess_list, kl_sum = \
                make_guess(Norm, self.grid, nfill=hyperparameters["nwalk"])
            # Load grid information
            cata_hist_data = self.grid.load_catalog_hist()
            # Perform fit
            Norm.fit_random_walk(
                                 fit_method = fit_method,
                                 nwalk = hyperparameters["nwalk"],
                                 nstep = hyperparameters["nstep"],
                                 kl_bins = self.grid.kl_bins,
                                 kl_sensitivity = self.grid.kl_sensitivity,
                                 fit_samples = self.grid.fit_samples,
                                 hist_data = cata_hist_data,
                                 coords = self.grid.coords,
                                 guess_list = guess_list,
                                 carryover = hyperparameters["carryover"],
                                 freeze_sensitivity = hyperparameters["freeze_sensitivity"],
                                )

        ## Emcee ##
        elif fit_method == "emcee":
            # load catalog samples
            posterior, prior = self.grid.load_catalog_samples()
            # Generate inverse prior
            inv_prior = np.power(prior, -1.)
            # Make intelligent guesses
            guess_list, kl_sum = \
                make_guess(Norm, self.grid)
            # Load grid information
            cata_hist_data = self.grid.load_catalog_hist()
            # Perform fit
            Norm.fit_emcee(
                           posterior, inv_prior,
                           fit_method = fit_method,
                           nwalk = hyperparameters["nwalk"],
                           nstep = hyperparameters["nstep"],
                           coords = self.grid.coords,
                           kl_bins = self.grid.kl_bins,
                           kl_sensitivity = self.grid.kl_sensitivity,
                           fit_samples = self.grid.fit_samples,
                           hist_data = cata_hist_data,
                           guess_list = guess_list,
                          )

        """

        # End time
        t_end = time.time()
        fit_time = t_end - t_start
        hyperparameters["fit_time"] = fit_time
        
        # Inform the user about the fit
        if self.verbose:
            seconds = int(fit_time % 60)
            minutes = int((fit_time / 60) % 60)
            hours = int(fit_time/3600)
            print("\nTime: %d hours, %d minutes, %d seconds"%(hours, minutes, seconds))

            # Calculate kl divergence
            this_kl = self.kl_divergence(**hyperparameters)
            print("kl: ", this_kl)
            
            # Load mean and cov for fit
            mean_cur, cov_cur = self.norm.read_physical()
            print("mean", mean_cur)
            print("cov: ")
            print(cov_cur)

            # Calculate standard deviation
            std_dev = np.sqrt(np.diag(cov_cur))
            print("std_dev: ", std_dev)

            # Calculate corelation params
            cor = self.norm.cor_of_params(self.norm.read_guess())
            print("Correlation:")
            print(cor)

        # Save simple fit
        if self.verbose:
            print("Saving fit")

        self.save_fit(
                      fit_method,
                      save_rule=save_rule,
                      **hyperparameters
                     )

    @staticmethod
    def load_fit(
                 fname_nal,
                 group,
                 event,
                 coord_tag,
                 fit_method,
                ):
        '''\
        Load an existing fit
        '''
        # Check if file exists
        if not isfile(fname_nal):
            raise RuntimeError("No such file: %s"%fname_nal)
        # Check if event exists
        db_init = Database(fname_nal)
        if not event in db_init.list_items():
            raise ValueError("No fit available for %s in %s"%(event,fname_nal))
        # Generate label
        label = "%s:%s:%s"%(coord_tag,group,fit_method)
        if not label in db_init.list_items(event):
            raise ValueError("No fit available for %s/%s in %s"%(event,label,fname_nal))
        fit_addr = join(event, label)
        #print("loading existing fit for %s"%label)
        db = Database(fname_nal, fit_addr)
        
        # Load coords
        coords = coord_tags[coord_tag]
        # Load scale
        scale = db.dset_value("scale")
        ndim = len(scale)
        # load mean and cov
        mean = db.dset_value("mean")/scale
        cov = db.dset_value("cov")/np.outer(scale,scale)
        # Load limits
        limits = db.dset_value("limits")

        # Initialize list of variables
        variables = []
        for i in range(ndim):
            # We use this a couple times
            coord = coords[i]
            # Initialize a parameter
            p = Parameter(
                          coord,
                          mean[i]*scale[i],
                          limits[i],
                          label=coord_labels[coord],
                         )
            # Append parameter
            variables.append(p)

        # Initialize normal object
        Norm = MultivariateNormal(variables, scale)
        # Convert params
        params = params_of_mu_cov(mean,cov)

        # When saving fits, you get to choose allow_singular.
        # When loading, you don't get to choose
        Norm.assign_guess(params)
        #print("%s %s successfully loaded!"%(event, label))

        return Norm

    def update_fit(
                   self,
                   fit_method,
                  ):
        '''\
        Load an existing fit
        '''
        # Generate label
        label = "%s:%s:%s"%(self.coord_tag,self.group,fit_method)
        fit_addr = join(self.event, label)
        print("loading existing fit for %s"%label)

        # Print fields
        fit_dsets = self.db.list_items(fit_addr, kind="dset")
        fit_attrs = self.db.attr_list(fit_addr)
        print(fit_dsets)
        print(fit_attrs)

        # Load the scale
        try:
            scale = self.db.dset_value(join(fit_addr, "scale"))
        except:
            scale = self.load_scale()
        # load mean and cov
        try:
            mean = self.db.dset_value(join(fit_addr, "mean"))/scale
            cov = self.db.dset_value(join(fit_addr, "cov"))/np.outer(scale,scale)
        except:
            print("update fit skipping. mean or covariance is left out.")
            return

        # Load coordinates
        if "coords" in fit_attrs:
            coords = self.db.attr_value(fit_addr, "coords")
        else:
            coords = coord_tags[self.coord_tag]
            self.db.attr_set(fit_addr, "coords", coords)

        # Check status of coord_tag
        if not "coord_tag" in fit_attrs:
            self.db.attr_set(fit_addr, "coord_tag", self.coord_tag)

        # Check event
        if not "event" in fit_attrs:
            self.db.attr_set(fit_addr, "event", self.event)

        # Check group
        if not "group" in fit_attrs:
            self.db.attr_set(fit_addr, "group", self.group)

        # Check limits
        if not "limits" in fit_dsets:
            self.db.dset_set(join(fit_addr, "limits"), self.limits)
        # Load limits
        limits = self.db.dset_value(join(fit_addr, "limits"))
                
        # Initialize list of variables
        #variables = self.load_variables()
        variables = []
        for i in range(self.ndim):
            # We use this a couple times
            coord = coords[i]
            # Initialize a parameter
            p = Parameter(
                          coord,
                          mean[i] * scale[i],
                          limits[i],
                          label=coord_labels[coord],
                         )
            # Append parameter
            variables.append(p)

        # Load scale
        if not "scale" in fit_dsets:
            # Load scale from grid
            scale = self.load_scale()
            # Save scale 
            self.db.dset_set(join(fit_addr, "scale"), scale)
        # Load scale
        scale = self.db.dset_value(join(fit_addr, "scale"))

        # Initialize normal object
        Norm = MultivariateNormal(variables, scale, random_state=self.rs)

        # Reorganize the parameters
        params = Norm.params_of_mu_cov(mean,cov)
        # Load the guess
        Norm.assign_guess(params)
        print("%s %s successfully loaded!"%(self.event, label))

        return


    @staticmethod
    def plot_likelihood(
                        savename,
                        fname_nal,
                        fname_mesh,
                        group,
                        event,
                        coord_tag,
                        fit_method,
                        verbose = False,
                        evaluation_res=100,
                       ):
        '''\
        Generate a likelihood corner plot for a fit
        '''
        from .density import Mesh
        from .plots.likelihood_corner import corner_cross_sample_normal
        import time
        # Common sense
        extension = savename.split('.')[-1]
        # Generate labels
        label = "%s:%s:%s"%(coord_tag,group,fit_method)
        fit_addr = join(event, label)
        mesh_label = "%s/%s:%s"%(event,coord_tag,group)
        # Load the fit
        Norm = NAL.load_fit(fname_nal, group, event, coord_tag, fit_method)
        # Load the mesh
        mesh = Mesh.load(fname_mesh,mesh_label)
        # Plot the fit
        if verbose:
            print("Generating corner plot for %s"%label)
            print("This may take some time.")
        # Make plot
        corner_cross_sample_normal(
                                   savename,
                                   Norm,
                                   mesh,
                                   label,
                                   extension=extension,
                                   title=event,
                                   verbose=verbose,
                                   evaluation_res=evaluation_res,
                                  )
        if verbose:
            print("Corner plot completed!")

    def generate_guesses(self, **kl_args):
        '''\
        Generate guesses for the fits!
        '''
        # Initialize guess list
        Xg = []

        # Initial guess
        Xg.append(self.norm.read_guess())

        # Simple fit guess
        self.fit_simple(save_rule="never",asign=False)
        Xg.append(self.norm.read_guess())
        simple_guess = self.norm.read_guess()

        # hist1d guesses
        mesh_guesses = self.norm.fit_mesh(self.mesh)
        for item in mesh_guesses:
            Xg.append(item)

        # Event fits
        event_labels = self.db.list_items(self.event)
        for label in event_labels:
            # Identify label information
            label_coord_tag, label_group, label_fit_method = \
                label.split(":")
            label_coords = coord_tags[label_coord_tag]
            # Try to identify visitor fit
            try:
               label_norm = NAL.load_fit(
                                         self.fname_nal,
                                         label_group,
                                         self.event,
                                         label_coord_tag,
                                         label_fit_method,
                                        )
            except:
                label_norm is None
            
            if not (label_norm is None):
                # Create a parameter map
                p_map = {}
                for i, item in enumerate(self.coords):
                    p_map[i] = None
                    for j, jtem in enumerate(label_coords):
                        if item == jtem:
                            p_map[i] = j

                # Success! Identify label params
                label_guess = np.copy(simple_guess)
                label_X = label_norm.read_guess()
                # Identify local quantities
                mu = mu_of_params(label_guess)[0]
                std = std_of_params(label_guess)[0]
                cor = cor_of_params(label_guess)[0]
                scale = self.norm.scale
                # Identify label quantities
                label_mu = mu_of_params(label_X)[0]
                label_std = std_of_params(label_X)[0]
                label_cor = cor_of_params(label_X)[0]
                label_scale = label_norm.scale
                # Loop through each coordinate in Norm's coordinate system
                for i in range(self.ndim):
                    # Do nothing if coordinate isn't in Vis's coordinate system
                    if not (p_map[i] is None):
                        # Identify scale ratio
                        scale_ratio = label_scale[p_map[i]]/scale[i]
                        # Update mu and std
                        mu[i] = label_mu[p_map[i]] * scale_ratio
                        std[i] = label_std[p_map[i]] * scale_ratio
                        # Update correlation factors
                        for j in range(i):
                            if not (p_map[j]) is None:
                                cor[i,j] = cor[j,i] = label_cor[p_map[i], p_map[j]]

                # Get params back
                cov = cov_of_std_cor(std,cor)[0]
                # Update guess
                label_guess = params_of_mu_cov(mu,cov)[0]
                # Update guesses
                Xg.append(label_guess)

                # Create an alternate guess with the simple corelation
                cor = cor_of_params(simple_guess)[0]
                # Get params back
                cov = cov_of_std_cor(std,cor)[0]
                # Update guess
                label_guess = params_of_mu_cov(mu,cov)[0]
                # Update guesses
                Xg.append(label_guess)
                    

        # Sort guess list
        guess_array = np.asarray(Xg)
        keep = self.norm.satisfies_constraints(guess_array)
        guess_array = guess_array[keep]
        Xg = []
        kl = []
        for i in range(guess_array.shape[0]):
            try:
                item_kl = self.norm.mesh_kl(
                                            self.mesh,
                                            guess_array[i],
                                            mode="parameter",
                                            **kl_args
                                           )[0]
                Xg.append(guess_array[i])
                kl.append(item_kl)
                if self.verbose:
                    print("guess pass",i, np.sum(item_kl), guess_array[i])
            except:
                if self.verbose:
                    print("guess fail",i, guess_array[i])

        # Mix and match
        kl_tmp = np.asarray(kl)
        guess_mix = np.copy(simple_guess)
        guess_ind = np.zeros(len(self.norm._parameters),dtype=int)
        for j in range(len(guess_mix)):
            guess_ind[j] = np.argmin(kl_tmp[:,j])
            guess_mix[j] = Xg[guess_ind[j]][j]
        
        # Try the mixmatch guess
        try:
            item_kl = self.norm.mesh_kl(
                                        self.mesh,
                                        guess_array[i],
                                        mode="parameter",
                                        **kl_args
                                       )[0]
            Xg.append(guess_array[i])
            kl.append(item_kl)
            if self.verbose:
                print("mixmatch guess pass", np.sum(item_kl))
                print("mixmatch ids", guess_ind)
        except:
            if self.verbose:
                print("guess fail", guess_mix)

        # Create an array of guesses for real this time
        Xg = np.asarray(Xg)
        Xg = self.norm.check_sample(Xg)
        # Make kl an array
        kl = np.asarray(kl)
        # Create the kl sum
        kl_sum = np.sum(kl[:,:self.ndim],axis=1) + \
                 np.sum(kl[:,2*self.ndim:],axis=1)
        # Sort the guess list
        index = np.argsort(kl_sum)
        Xg = Xg[index]
        kl_sum = kl_sum[index]

        return Xg, kl_sum



#!/usr/bin/env python3
'''\
Test the Multivariate Normal object and methods
'''
######## Imports ########
import numpy as np
from scipy import stats
from gwalk.model import Parameter
from gwalk.bounded_multivariate_normal import MultivariateNormal
from gwalk.utils.multivariate_normal import params_of_norm_mu_cov
from gwalk.utils.multivariate_normal import params_reduce_1d, params_reduce_2d
from gwalk.utils.multivariate_normal import multivariate_normal_marginal1d
from gwalk.utils.multivariate_normal import multivariate_normal_marginal2d
from gwalk.utils.multivariate_normal import cov_of_std_cor
from gwalk.utils.multivariate_normal import multivariate_normal_pdf as gwalk_norm
from gwalk.utils.multivariate_normal import analytic_enclosed_integral
from gwalk.multivariate_normal import pdf as new_pdf
from scipy.stats import multivariate_normal as scipy_norm
from gwalk.density import Mesh
from gp_api.utils import sample_hypercube
import time

######## Settings ########

limits = np.asarray([
                     [0., 1.],
                     [0., 1.],
                     [0., 1.],
                    ])

norm = 0.
mu = np.asarray([0.5,0.95,0.5])
print("mu:")
print(mu)
sig = np.ones_like(mu)*0.1
scale = np.copy(sig)
rho_ij = 0.0
rho = np.asarray([
                  [1., rho_ij, rho_ij,],
                  [rho_ij, 1., rho_ij,],
                  [rho_ij, rho_ij, 1.,],
                 ])
Sig = cov_of_std_cor(sig, rho)[0]
print("Covariance:")
print(Sig)

ndim = mu.size
max_bins=10

n_rvs = int(1e6)
n_max = 3
kl_sensitivity = None

seed = 0
rs = np.random.RandomState(seed)

######## Test ########

params = []
for i in range(ndim):
    params.append(Parameter("p%d"%i,mu[i],limits[i]))

######## Build MultivariateNormal ########
# Initialize relation
MV = MultivariateNormal(params, scale=scale,seed=0)


Ys = scipy_norm.rvs(mean=mu,cov=Sig,size=n_rvs,random_state=rs)
Xparams = params_of_norm_mu_cov(norm, mu, Sig).flatten()
Xparams_scaled = np.copy(Xparams)
Xparams_scaled[1:ndim+1] /= scale
Xparams_scaled[ndim+1:2*ndim+1] /= scale
MV.assign_guess(Xparams_scaled)
_mu, _cov = MV.read_physical()
assert np.allclose(_mu, mu)
assert np.allclose(_cov, Sig)
#Ys = norm.summands[0].sample_uniform(n_rvs)

print("Testing accuracy and timing")
t0 = time.time()
lnL0 = np.log(scipy_norm.pdf(
    Ys, mean=mu, cov=Sig))
t1 = time.time()
L1 = gwalk_norm(Xparams,Ys,log_scale=False,scale=scale).flatten()
lnL1 = np.log(L1)
t2 = time.time()
lnL2 = gwalk_norm(Xparams,Ys,log_scale=True).flatten()
t3 = time.time() 
lnL3 = np.log(MV.likelihood(Ys).flatten())
t4 = time.time()
lnL4 = MV.likelihood(Ys,log_scale=True).flatten()
t5 = time.time()
lnL5 = np.log(new_pdf(mu, Sig, Ys, scale=scale))
t6 = time.time()
lnL6 = new_pdf(mu, Sig, Ys, scale=scale, log_scale=True)
t7 = time.time()
keep = np.isfinite(lnL3)
print("Scipy time:", t1-t0)
print("gwalk function time:", t2-t1)
print("gwalk function (log) time:", t3-t2)
print("gwalk object time:", t4-t3)
print("gwalk object time(log):", t5-t4)
print("new time:", t6-t5)
print("new log time:", t7-t6)
print("gwalk function maximum error:",
    np.max(np.abs(lnL1[keep] -lnL0[keep])))
print("gwalk function (log) maximum error:",
    np.max(np.abs(lnL2[keep] -lnL0[keep])))
print("gwalk object maximum error:",
    np.max(np.abs(lnL3[keep] -lnL0[keep])))
print("gwalk object (log) maximum error:",
    np.max(np.abs(lnL4[keep] -lnL0[keep])))
print(lnL5)
print("New maximum error:",
    np.max(np.abs(lnL5[keep] - lnL0[keep])))
print("New log maximum error:",
    np.max(np.abs(lnL6[keep] - lnL0[keep])))

raise Exception

# Calculate integral
print("Sample integral over enclosed space:")
I = np.ones_like(np.exp(lnL1),dtype=bool)
for i in range(ndim):
    I &= Ys[:,i] > limits[i][0]
    I &= Ys[:,i] < limits[i][1]
print(np.sum(I)/lnL1.size)
print("Analytic integral over enclosed space:")
I_anal = float(MV.analytic_enclosed_integral(assign=True))
print(I_anal)

As = np.argsort(np.exp(lnL1).flatten())

Ym = Ys[As[-n_max:]]
Pm = np.exp(lnL1)[As[-n_max:]]

Xg = MV.read_guess()
MV.fit_simple(Ys,assign=True)
Xg = MV.read_guess()

Xs = np.empty((n_max,Xg.size))
for i in range(Xg.size):
    if i < n_max:
        Xs[:,i] = Ym[:,i]
    else:
        Xs[:,i] = Xg.flatten()[i]

Px = MV.likelihood(Ys,X=Xs)

######## Test saving ########

MV.save("test_foo.hdf5","mylabel")
n2 = MultivariateNormal.load("test_foo.hdf5","mylabel",normalize=True)
print(n2.read_guess())
P2 = n2.likelihood(Ys,X=Xs) * I_anal
print(np.max(P2-Px))



#!/usr/bin/env python3
'''\
Build gwalk-friendly catalog database for samples from the GWTC catalogs

build_catalog.py

'''
######## Imports ########
from gwalk.catalog import Catalog
import sys

######## Arguments ########
print(sys.argv)
if len(sys.argv) == 4:
    # Name of the catalog file
    fname_db =      sys.argv[1]
    # Name of GW release
    release =       sys.argv[2]
    # Location of GWXXXXXX.h5 files from catalog release
    release_dir =   sys.argv[3]

    # Initialize catalog
    cata = Catalog(fname_db)
    # Build release
    cata.build_catalog(release, release_dir)

elif (len(sys.argv) == 3) and (sys.argv[2] == "clean"):
    # Name of the catalog file
    fname_db =      sys.argv[1]
    # Create a new database
    cata = Catalog(fname_db, clean=True)

else:
    raise RuntimeError("Usage: python3 build_catalog.py [fname_db] [release] [release directory]")




#!/usr/bin/env python3
'''\
Test the Mesh object fitting, saving, and loading
'''

def main():
    #### Imports ####
    import time
    import numpy as np
    from gwalk.tools.fit_likelihood_grid import fit_grid
    from gwalk.utils.multivariate_normal import ndim_of_n_param
    from gwalk.utils.multivariate_normal import n_param_of_ndim
    from gwalk.utils.multivariate_normal import params_of_norm_mu_cov
    from gwalk.utils.multivariate_normal import multivariate_normal_pdf

    #### Settings ####

    limits = np.asarray([
                         [0., 1.],
                         [0., 1.],
                         [0., 1.],
                        ])

    mu = np.asarray([0.5,0.95,0.5])
    ndim = mu.size
    sig = np.ones_like(mu)*0.1
    norm = 5.

    n_rvs = int(1e5)
    nselect = 1000
    #nselect=n_rvs
    seed = 0
    rs = np.random.RandomState(seed)
    filename = "test_mesh.hdf5"
    objective = "kl"
    #objective = "lstsq"
    method="L-BFGS-B"

    #### Generate params ####
    params = np.zeros((1,n_param_of_ndim(ndim)))
    params[0,0] = norm
    params[0,1:ndim+1] = mu
    params[0,ndim+1:2*ndim+1] = sig

    #### run code ####

    ## random samples ##
    samples = (rs.uniform(size=(n_rvs,ndim))*(limits[:,1]-limits[:,0])) + limits[:,0]
    ## pdf values ##
    values = multivariate_normal_pdf(params, samples, log_scale=True).flatten() + norm

    ## Downselect samples ##
    select = np.argsort(values)[-nselect:]
    samples = samples[select]
    values = values[select]

    ## Fit grid ##
    t0 = time.time()
    grid, MV = fit_grid(
                        samples,
                        values,
                        limits,
                        seed=seed,
                        p_labels=None,
                        objective=objective,
                        method=method,
                       )
    t1 = time.time()


    Xg = MV.read_guess()
    ## Print parameter values ##
    params[:,1:ndim+1] /= MV.scale
    params[:,ndim+1:2*ndim+1] /= MV.scale
    print("fit time: %f seconds"%(t1-t0))
    print("Objective: %s"%(objective))
    print("--------")
    print("True params:")
    print(params)
    print("kl divergence with true params:")
    print(grid.nal_kl_div(MV,params))
    print("--------")
    print("Simple params:")
    print(grid.X_simple)
    print("kl divergence with simple params:")
    print(grid.nal_kl_div(MV,grid.X_simple))
    print("ln(simple error:)")
    print(np.log(np.abs(grid.X_simple - params)))
    print("--------")
    print("Optimized params")
    print(MV.read_guess())
    print("kl divergence with optimized params:")
    print(grid.nal_kl_div(MV,Xg))
    print("ln(optimized error:)")
    print(np.log(np.abs(Xg - params)))
    return

######## Execution ########
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
'''\
Test the Mesh object fitting, saving, and loading
'''

######## Imports ########

import numpy as np
from gwalk.model.relation import Relation
from gwalk.model.parameter import Parameter
from gwalk.model.covariance_relation import Covariance

######## Settings ########

limits = np.asarray([
                     [0., 1.],
                     [0., 1.],
                     [0., 1.],
                    ])

mu = np.asarray([0.5,0.9,0.5])
ndim = mu.size
max_bins=10

n_rvs = int(1e6)

seed = 0
rs = np.random.RandomState(seed)

######## Test ########

cov = Covariance(ndim, random_state=rs)

cov_values = cov.read_cov()
guess = cov.read_guess()

print(cov_values)
print(cov.read_guess())
print(cov.cov_of_params(guess))
print(cov.satisfies_constraints(guess))



#!/home/xevra/.local/bin/python3
'''\
Keep the data together so it doesn't need to be recalculated every time
'''

######## Imports ########
import numpy as np
import os
from os.path import join, isdir, isfile
from scipy.stats import gaussian_kde
from fast_histogram import histogram1d, histogram2d
from .data import Database
from .coordinates import coord_labels as COORD_LABELS
from .coordinates import coord_tags as COORD_TAGS
from .catalog import Catalog
#from .fit.hist import get_bin_ids, hist1d_from_ids, hist2d_from_ids

######## GRID Object ########

class Grid(object):
    '''\
    Catalog object

    Inputs:
        TODO
    methods:
        TODO
    '''
    def __init__(
                 self,
                 wkdir,
                 event,
                 group,
                 coord_tag,
                 clean=False,
                 seed=20210802,
                 convolve=None,
                ):
        '''\
        Initialize Grid object
        '''
        print("Initializing grid for %s %s %s"%(event, group, coord_tag))
        # Store information
        self.wkdir = wkdir
        self.event = event
        self.group = group
        self.coord_tag = coord_tag
        self.coords = COORD_TAGS[coord_tag]
        self.prior = self.coords[-1]

        # Identify file name
        self.fname = join(wkdir, "likelihood_grid.hdf5")
        # Initialize catalog
        self.fname_cata = join(wkdir,"catalog_samples.hdf5")
        self.cata = Catalog(self.fname_cata)
        # find fname_release
        self.release = self.cata.release_of_event(event)
        self.fname_release = join(wkdir, "%s.nal.hdf5"%self.release)
        # Clean file
        if clean and isfile(self.fname):
            os.system("rm %s"%self.fname)

        # Update group address
        self.addr = join(self.release, event, self.group, self.prior)

        # Open database
        self.db = Database(self.fname)
        # Initialize release
        if not self.db.exists(self.release,kind='group'):
            self.db.create_group(self.release)
        # Initialize event
        if not self.db.exists(join(self.release, event), kind='group'):
            self.db.create_group(join(self.release,event))
        # Initialize group
        if not self.db.exists(self.addr):
            self.db.create_group(self.addr)
        self.db = Database(self.fname, group=self.addr)

        # Initialize random state
        self.rs = np.random.RandomState(seed)
        self.seed = seed

        # Load coordinates
        self.ndim = len(self.coords) - 1

        # Attempt to load attributes
        self.load_attrs(require=False)
        self.load_limits(require=False)
        print("Grid initialized!")

        # Check convolve
        if not (convolve is None):
            if not type(convolve) == int:
                raise TypeError("Convolve must be None, or an int")
            elif convolve < 1:
                raise RuntimeError("Convolve must be > 1")
        self.convolve = convolve

    #### I/O ####

    def check_status(self):
        '''\
        Check on the status of a grid
        '''
        # Check attributes
        attrs = self.db.attr_dict(".")
        try:
           assert "fit_samples" in attrs
           assert "kl_bins" in attrs
           assert "kl_samples" in attrs
           assert "kl_sensitivity" in attrs
           assert "seed" in attrs
        except: 
            print("grid: attrs are not complete")
            print(attrs)
            return False

        # Check datasets
        dsets = self.db.list_items()
        try:
            assert "limits_%s"%(self.coord_tag) in dsets
            assert "edges_%s"%(self.coord_tag) in dsets
            assert "centers_%s"%(self.coord_tag) in dsets
            #assert "kde_samples" in dsets
            #assert "kde_sample_pdf" in dsets
        except:
            print("grid: fixed dsets are not complete")
            print(dsets)
            return False

        # Check more dsets
        try:
            # Loop through each dimension
            for i in range(self.ndim):
                assert "catalog_hist1d_%s"%self.coords[i] in dsets

                # Loop through other dimensions
                for j in range(i):
                    # Check 2-D histogram
                    assert "catalog_hist2d_%s_%s"%(self.coords[i],self.coords[j]) in dsets
        except:
            print("grid: coordinate dsets are not complete")
            print(dsets)
            return False

        return True

    def list_all(self):
        '''\
        list all the attrs and datasets
        '''
        # Print attributes
        attrs = self.db.attr_dict('.')
        for item in attrs:
            print(item, attrs[item])
        # List databases
        dsets = self.db.list_items()
        for item in dsets:
            print(item)

    def load_catalog_samples(
                             self,
                             mirror = False,
                            ):
        '''\
        Load posterior samples from the catalog file
        '''
        # Extract the data
        pdict = self.cata.load_data(
                                    self.event,
                                    self.group,
                                    COORD_TAGS[self.coord_tag]
                                   )
        # Stack samples
        posterior = []
        for item in pdict:
            if item.startswith("prior"):
                prior = np.asarray(pdict[item])
            else:
                posterior.append(pdict[item])
        posterior = np.asarray(posterior).T

        if mirror:
            # Join the mirrored posterior samples with the unmirrored ones
            # Create a second set of posterior samples, mirrored about eta = 0.25
            posterior_mirror = posterior.copy()
            posterior_mirror[:,1] = 0.5 - posterior[:,1]
            # Append those samples with the ordinary samples
            posterior_augment = np.append(posterior, posterior_mirror, axis=0)
            prior_augment = np.append(prior, prior)
            return posterior_augment, prior_augment
        else:
            return posterior, prior

    def load_KDE(
                 self,
                 mirror = False,
                ):
        '''\
        Load the KDE using the posterior sample catalog
        '''
        ## Generate Initial KDE ##
        # Load samples
        posterior, prior = \
                self.load_catalog_samples(
                                          mirror=False
                                         )
        # Update weights
        inv_prior = np.power(prior, -1.)
        # Generate KDE
        K_init = gaussian_kde(posterior.T, weights=inv_prior)

        # Handle unmirrored samples
        if mirror == False:
            return K_init

        ## Generate Mirrored KDE ##

        # Load samples
        posterior, prior = \
                self.load_catalog_samples(
                                          mirror=True
                                         )
        # Update weights
        inv_prior = np.power(prior, -1.)

        # Fit KDE
        self.K = gaussian_kde(posterior.T, weights=inv_prior)

        # identify scale factor
        #K_scale = (K_init.factor/self.K.factor)**2
        # Update precision matrix
        self.K.inv_cov = K_init.inv_cov# * K_scale

        return self.K


    def load_attrs(self,require=True):
        '''\
        Attempt to load attributes
        '''
        if self.check_status():
            self.__dict__.update(self.db.attr_dict("."))
            return self.db.attr_dict(".")
        elif require:
            raise RuntimeError("Grid has not been initialized properly")

    def load_limits(self,require=True):
        '''\
        Attempt to load limits
        '''
        if self.db.exists("limits_%s"%(self.coord_tag)):
            self.limits = self.db.dset_value("limits_%s"%(self.coord_tag))
            return self.limits
        elif require:
            raise RuntimeError("Grid: Limits not found.")

    def load_catalog_hist(self):
        '''\
        Attempt to load catalog histogram
        '''
        # Check convolve
        if not (self.convolve is None):
            # Import convolve
            from scipy.signal import convolve
            # Generate convolve stamp
            cstamp = np.ones((self.convolve,self.convolve))
        # Attempt to load data
        try:
            hist_data = {}
            hist_data["limits"] = self.limits
            hist_data["edges"] = self.db.dset_value("edges_%s"%(self.coord_tag))
            hist_data["centers"] = self.db.dset_value("centers_%s"%(self.coord_tag))
            for i in range(self.ndim):
                hist_data["hist1d_%s"%self.coords[i]] = \
                    self.db.dset_value("catalog_hist1d_%s"%(self.coords[i]))
                for j in range(i):
                    hist_tag = "hist2d_%s_%s"%(self.coords[i],self.coords[j])
                    hist_data[hist_tag] = \
                        self.db.dset_value("catalog_hist2d_%s_%s"%(
                                self.coords[i], self.coords[j]))
                    if not (self.convolve is None):
                        hist_data[hist_tag] = convolve(
                                                       hist_data[hist_tag],
                                                       cstamp,
                                                       mode="same",
                                                      )
                    center_i, center_j = np.meshgrid(
                                                     hist_data["centers"][i],
                                                     hist_data["centers"][j],
                                                    )
                    hist_data["mesh2d_%s_%s"%(
                                              self.coords[i],
                                              self.coords[j],
                                             )] = \
                                np.asarray([
                                            center_i.flatten(),
                                            center_j.flatten(),
                                           ]).T

        except:
            raise RuntimeError("Failed to load catalog histogram data")

        return hist_data

    #### Build ####

    def select_limits(
                      self,
                      posterior,
                     ):
        '''\
        Select reasonable limits for the grid
        '''
        self.limits = np.zeros((self.ndim, 2))
        for i in range(self.ndim):
            soft_limits = np.asarray([np.min(posterior[:,i]), np.max(posterior[:,i])])
            '''
            x = posterior[:,i]
            sample_mean = np.mean(x)
            sample_var = np.std(x)
            sig = 1.26*sample_var
            p_limits = np.percentile(x,[1,99])
            #print(soft_limits, sample_mean, sample_var)
            if p_limits[0] < sample_mean - sig:
                self.limits[i,0] = p_limits[0]
            else:
                self.limits[i,0] = soft_limits[0]
            if p_limits[1] > sample_mean + sig:
                self.limits[i,1] = p_limits[1]
            else:
                self.limits[i,1] = soft_limits[1]
            print(self.limits[i] == p_limits,self.limits[i])
            '''
            self.limits[i] = soft_limits
        self.db.dset_set("limits_%s"%self.coord_tag, self.limits)

    def build_catalog_histogram(
                                self,
                                posterior,
                                inv_prior,
                                kl_bins,
                               ):
        '''\
        Build the histogram for catalog samples
        '''
        # Loop through each dimension
        for i in range(self.ndim):
            # Generate a 1-D histogram in that dimension
            if not self.db.exists("catalog_hist1d_%s"%self.coords[i]):
                print("constructing catalog_hist1d_%s"%self.coords[i])
                print(self.limits[i])
                hist1d = histogram1d(posterior[:,i], range=self.limits[i], bins=kl_bins, weights=inv_prior)
                # Save the 1-D histogram
                self.db.dset_set("catalog_hist1d_%s"%self.coords[i], hist1d)
            # Loop through other dimensions
            for j in range(i):
                # Generate 2-D histogram
                if not self.db.exists("catalog_hist2d_%s_%s"%(self.coords[i],self.coords[j])):
                    print("constructing catalog_hist2d_%s_%s"%(self.coords[i],self.coords[j]))
                    hist2d = histogram2d(
                                         posterior[:,i], posterior[:,j],
                                         range= [self.limits[i],self.limits[j]],
                                         bins=kl_bins,
                                         weights=inv_prior,
                                        )
                    # Save 2-D histogram
                    self.db.dset_set("catalog_hist2d_%s_%s"%(self.coords[i],self.coords[j]),hist2d)

    def build_grid(self, fit_samples, kl_bins, kl_samples, kl_sensitivity):
        '''\
        Generate the samples and histograms for a model
        '''
        print("Building a grid for %s %s %s"%(self.event, self.group, self.coord_tag))
        print("This may take some time")
        print("grid parameters: ")

        ## Write attributes ##
        kl_bins = 100
        grid_attrs = {
                      "fit_samples"     : fit_samples,
                      "kl_bins"         : kl_bins,
                      "kl_samples"      : kl_samples,
                      "kl_sensitivity"  : kl_sensitivity,
                      "seed"            : self.seed,
                     }
        print(grid_attrs)
        self.db.attr_set_dict('.', grid_attrs)

        print("Generating catalog sample posterior and prior data")
        ## Load non-mirrored posterior samples ##
        # load data
        posterior, prior = self.load_catalog_samples()
        print("Success!")
        # Update weights
        print("Inverting prior!")
        inv_prior = np.power(prior, -1.)
        print("Success!")
        
        ## Generate Limits ##
        print("Constructing limits!")
        self.select_limits(posterior)
        print("Success!")

        ## Check limits for singular values ##
        if (self.limits[:,1] == self.limits[:,0]).any():
            print("Singular grid limits found! Aborting")
            return 1

        ## Construct edges and centers ##
        edges = np.empty((self.ndim,kl_bins + 1))
        centers = np.empty((self.ndim, kl_bins))
        for i in range(self.ndim):
            edges[i] = np.linspace(self.limits[i][0],self.limits[i][1],kl_bins+1)
            centers[i] = 0.5*(edges[i,1:] + edges[i,:-1])
        self.db.dset_set("edges_%s"%self.coord_tag, edges)
        self.db.dset_set("centers_%s"%self.coord_tag, centers)

        ## Generate Posterior histogram ##
        print("Constructing Catalog histogram")
        self.build_catalog_histogram(posterior, inv_prior, kl_bins)
        print("Success!")
        return 0



#!/usr/bin/env python3
'''\
Test the Mesh object fitting, saving, and loading
'''

######## Imports ########

import numpy as np
from gwalk.model.relation import Relation
from gwalk.model.parameter import Parameter

######## Settings ########

limits = np.asarray([
                     [0., 1.],
                     [0., 1.],
                     [0., 1.],
                    ])

mu = np.asarray([0.5,0.9,0.5])
ndim = mu.size
max_bins=10

n_rvs = int(1e6)

seed = 0
rs = np.random.RandomState(seed)

######## Test ########

params = []
for i in range(ndim - 1):
    params.append(Parameter("p%d"%i,mu[i],limits[i]))

# Initialize relation
R1 = Relation(parameters=params)

# Sample R1
samples = R1.sample_uniform(n_rvs)
print(samples)

# Initialize another relation
R2 = Relation(parameters=Parameter("p%d"%(ndim-1),mu[-1],limits[-1]))

# Add 
A = R1 + R2

# Sample a hypercube
cube = A.hypercube(size=n_rvs,constrained=True)
print(cube)



#!/usr/bin/env python3
'''\
Test the Mesh object fitting, saving, and loading
'''
######## IMPORTS ########
import argparse
from os.path import isfile, join
import time
import numpy as np
from xdata import Database
from gwalk.multivariate_normal import random_gauss_params
from gwalk.optimize_likelihood_grid import optimize_likelihood_grid
from gwalk.multivariate_normal import pdf as multivariate_normal_pdf
from gwalk.multivariate_normal import params_of_offset_mu_cov
from gwalk.multivariate_normal import mu_cov_of_params
from gwalk.multivariate_normal import cor_of_cov
from gwalk.multivariate_normal import offset_of_params
from gwalk.multivariate_normal import MultivariateNormal

######## Argparse ########
def arg():
    parser = argparse.ArgumentParser()
    parser.add_argument("--fname-samples", required=True,
        help="file location for posterior samples")
    parser.add_argument("--fname-out", default=None,
        help="File location for MV")
    opts = parser.parse_args()
    assert isfile(opts.fname_samples)
    return opts

######## Settings ########

APPROXIMANT="C00:NRSur7dq4"
N_SELECT = 1000
OBJECTIVE = "lstsq"
METHOD="SLSQP"

N_RVS = int(1e6)
N_MAX = 3
HYPERCUBE_RES = 100
SEED = 0
RS = np.random.RandomState(SEED)
NDIM = 3
NSELECT = 1000
LIMITS = np.asarray([
                     [-7., 7.],
                     [-7., 7.],
                     [-7., 7.],
                    ])


######## Tests ########
def construct_GW231123(opts):
    #### Load data ####
    print(opts.fname_samples)
    db = Database(opts.fname_samples,group=APPROXIMANT)
    print(db.list_items())
    fields = db.dset_fields("posterior_samples")
    for key in fields:
        print(key)
    mc = db.dset_value("posterior_samples",field="chirp_mass_source")
    print(f"mc: {np.percentile(mc,[0,10,33,50,68,90,100])}")
    eta = db.dset_value("posterior_samples",field="symmetric_mass_ratio")
    print(f"eta: {np.percentile(eta,[0,10,33,50,68,90,100])}")
    chi_1 = db.dset_value("posterior_samples",field="a_1")
    print(f"chi_1: {np.percentile(chi_1,[0,10,33,50,68,90,100])}")
    chi_2 = db.dset_value("posterior_samples",field="a_2")
    print(f"chi_2: {np.percentile(chi_2,[0,10,33,50,68,90,100])}")
    chi_eff = db.dset_value("posterior_samples",field="a_1")
    print(f"chi_eff: {np.percentile(chi_eff,[0,10,33,50,68,90,100])}")
    chi_p = db.dset_value("posterior_samples",field="a_2")
    print(f"chi_p: {np.percentile(chi_p,[0,10,33,50,68,90,100])}")
    lnP = db.dset_value("posterior_samples",field="log_likelihood")
    print(f"lnP: {np.percentile(lnP,[0,10,33,50,68,90,100])}")
    print(f"Loaded {mc.size} posterior samples")
    samples = np.asarray([
        mc,
        eta,
        chi_1,
        chi_2,
    ]).T

    #### Generate a grid ####
    MVsimple = MultivariateNormal.from_samples(samples, seed=SEED)

    ## TODO KEEP ##

    ## Downselect samples ##
    #select = np.argsort(lnP)[-NSELECT:]
    #samples = samples[select]
    #lnP = lnP[select]
    print("Beginning optimization")

    ## Fit grid ##
    t0 = time.time()
    MV = optimize_likelihood_grid(
                                  samples,
                                  lnP=lnP,
                                  seed=SEED,
                                  objective=OBJECTIVE,
                                  method=METHOD,
                                 )
    t1 = time.time()


    Xg = MV.params
    ## Print parameter values ##
    print("fit time: %f seconds"%(t1-t0))
    print("Objective: %s"%(OBJECTIVE))
    print("--------")
    print("--------")
    print("Simple params:")
    print(MVsimple.params)
    print("--------")
    print("Optimized params")
    print(f"mu: {MV.mu}")
    print(f"std: {MV.std}")
    if opts.fname_out is not None:
        MVsimple.save(opts.fname_out,"GW231123/mc_eta_chi1_chi2:NRSur7dq4:simple")
        MV.save(opts.fname_out,"GW231123/mc_eta_chi1_chi2:NRSur7dq4:select")
    return

######## Main ########
def main(): 
    opts = arg()
    construct_GW231123(opts)
    return

######## Execution ########
if __name__ == "__main__":
    main()



from gwalk.multivariate_normal.pdf import *
from gwalk.multivariate_normal.decomposition import *
from gwalk.multivariate_normal.utils import *
from gwalk.multivariate_normal.object import MultivariateNormal



#!/usr/bin/env python3
'''\
Test the Mesh object fitting, saving, and loading
'''

######## Imports ########

import numpy as np
from gwalk.density import Mesh
from gwalk.utils.multivariate_normal import n_param_of_ndim
from gwalk.utils.multivariate_normal import params_of_norm_mu_cov

######## Settings ########

limits = np.asarray([
                     [0., 1.],
                     [0., 1.],
                     [0., 1.],
                    ])

mu = np.asarray([0.5,0.95,0.5])
sig = np.ones_like(mu)*0.1
ndim = mu.size
min_bins = 8
max_bins1d=100
max_bins2d=20
evaluation_res = 100
nwalk = 100

n_rvs = int(1e6)

seed = 0
rs = np.random.RandomState(seed)

filename = "test_mesh.hdf5"

#### Generate params ####
params = np.zeros((1,n_param_of_ndim(ndim)))
params[0,0] = 1.0
params[0,1:ndim+1] = mu
params[0,ndim+1:2*ndim+1] = sig

######## Generate Samples ########

# Initialize space for random samples
samples = rs.normal(loc=mu,size=(n_rvs,ndim),scale=sig)


######## Fit Mesh ########
mesh = Mesh.fit(
                samples,
                ndim,
                limits=limits,
                min_bins=min_bins,
                max_bins1d=max_bins1d,
                max_bins2d=max_bins2d,
               )

######## Serialization ########

# Save mesh
mesh.save(filename)

# Load mesh
loadmesh = Mesh.load(filename)

# Assert equal
assert mesh == loadmesh

######## Test norm functionality ########

print("generating evaluation set")
mesh.generate_evaluation_set(evaluation_res)
print("Constructing MV variable")
MV = mesh.construct_nal()
print("MV guess:",MV.read_guess())
print("Calling mesh fit to samples")
mesh.nal_fit_to_samples(MV,samples)
print("MV guess:",MV.read_guess())
print("calling mesh guesses")
Xg = mesh.nal_mesh_guesses(MV)
Xg = np.append(Xg,MV.check_sample(MV.read_guess()),axis=0)
#, axis=0)

#### test norm convergence
print("calling nal_kl_div")
kl = mesh.nal_kl_div(MV,Xg,mode='mean')
print("kl divergence for first guesses:")
print(kl)

print("calling init walkers")
Xg = mesh.nal_init_walkers(MV,nwalk,Xg=Xg)
print("Shape of long list of guesses:")
print(Xg.shape)
print("number of guesses which satisfy constraints:")
print(np.sum(MV.satisfies_constraints(Xg)))
kl = mesh.nal_kl_div(MV,Xg,mode='mean')
print("best kl for initial walkers:")
print(np.min(kl))
print("random walk!")
mesh.nal_fit_random_walk(MV,Xg)
Xgenetic = MV.read_guess()
print("Optimizaed parameters:")
print(Xgenetic)
print("kl for final parameters:")
print(mesh.nal_kl_div(MV,Xgenetic))
## Print parameter values ##
params[:,1:ndim+1] /= MV.scale
params[:,ndim+1:2*ndim+1] /= MV.scale
print("True parameters:")
print(params)
print("kl for true parameters:")
print(mesh.nal_kl_div(MV,params))



'''\
Correction functions for symmetric mass ratio cutoff
'''
import numpy as np
#### Globals ####
_INV_RT2 = 1./np.sqrt(2)
_LOG_2PI = np.log(2*np.pi)

def fast_eta_cdf(mu, sig, boundary=0.25):
    '''\
    Find the normalization factors for many mirrored guassians

    Inputs:
        mu: The expected value of eta
        sig: The variance in eta
        boundary: Where is data being mirrored?

    CDF = 1/2 * (1 + erf((x-mu)/(sig\sqrt(2))))
    '''
    from scipy.special import erf
    return 0.5 * (1 + erf(_INV_RT2*(boundary - mu)/sig))

def eta_correction_factor(cur, boundary=0.25):
    '''\
    Find the correction factor for the mirrored sample
    '''
    # Make at least 2d
    cur = np.atleast_2d(cur)
    # mu is eta
    mu = cur[:,1]
    # variance in eta
    sig = cur[:,4]
    
    xi = fast_eta_cdf(mu, sig, boundary=boundary)
    return 0.5/xi




#!/usr/bin/env python3
'''\
Generate makefiles
'''

######## Globals ########

RUN_ID = int("0001")
RUN_DIR = '/home/xevra/Event_Likelihood_Approximation/nal-runs/run_%s'%(str(RUN_ID).zfill(4))
RUN_DIR_0 = '/home/xevra/Event_Likelihood_Approximation/nal-runs/run_0000'
MAKE_DIR = '/home/xevra/Event_Likelihood_Approximation/gwalk/tools/NAL_makefiles'
#MAKE_GWALK = '/home/xevra/Event_Likelihood_Approximation/gwalk/tools/make_gwalk.py'
MAKE_GWALK = 'make_gwalk.py'
#METHODS = ["simple", "genetic", "select"]

COORD_TAG_RUNS = []
COORD_TAG_RUNS.append("aligned3d")
COORD_TAG_RUNS.append("aligned3d_source")
COORD_TAG_RUNS.append("aligned3d_dist")
COORD_TAG_RUNS.append("mass_tides")
COORD_TAG_RUNS.append("mass_tides_source")
COORD_TAG_RUNS.append("aligned_tides")
COORD_TAG_RUNS.append("aligned_tides_source")
COORD_TAG_RUNS.append("aligned_tides_dist")
COORD_TAG_RUNS.append("spin6d")
COORD_TAG_RUNS.append("precessing8d")
COORD_TAG_RUNS.append("precessing8d_source")
COORD_TAG_RUNS.append("precessing8d_dist")
COORD_TAG_RUNS.append("precessing_tides_source")
COORD_TAG_RUNS.append("full_precessing_tides")


######## Imports ########
import numpy as np
import os
from os.path import join, isfile, isdir
import sys
from contextlib import contextmanager
from gwalk import NAL
from gwalk.grid import Grid
from gwalk.catalog import Catalog
from gwalk.catalog import EVENTS
from gwalk.coordinates import coord_tags as COORD_TAGS

######## Initialize ########

# Name the database
fname_cata = join(RUN_DIR_0, "catalog_samples.hdf5")
# Build a fresh catalog
cata = Catalog(fname_cata)

######## Gather Release Waveforms ########

# Initialize waveforms
Events = {}
Fields = {}

for release in EVENTS:
    # Identify release
    groups = cata.groups_of_release(release)
    for group in groups:
        Events[group] = cata.events_of_group(group)
        Fields[group] = cata.fields_of_group(group)

######## Generate Makefiles ########

for group in Events:

    #### Setup General Lines ####
    # Replace escape characters
    events = Events[group]
    fields = Fields[group]
    coord_tags = cata.tags_of_group(group)
    group = group.replace(":","\:")
    # Generate file name
    fname_group = join(MAKE_DIR, "make_%s.mk"%(group))
    # Initialize lines
    lines = []
    lines.append("all:\n")
    lines.append("RUN_ID = %s\n"%(str(RUN_ID).zfill(4)))
    lines.append("RUN_DIR = %s\n"%(RUN_DIR))
    # generate event line
    event_line = "%s = "%(group)
    #if len(Events[group]) == 1:
    #    event_line = event_line + "%s None "%(event)
    #else:
    for event in events:
        event_line = event_line + "%s "%(event)
    event_line = event_line + '\n'
    lines.append(event_line)
    # generate Initialize lines
    build = "build:\n" + \
        "\trm -rf  ${HOME}/.local/lib/python*/site-packages/gwalk* \n" +\
	    "\tpip3 install --user ${HOME}/Event_Likelihood_Approximation/gwalk\n"
    lines.append(build)
    run_dir_job = "$(RUN_DIR)/catalog_samples.hdf5: build\n" +\
        "\tpython3 %s $(RUN_ID) initialize\n"%MAKE_GWALK +\
        "\tmkdir -p $(RUN_DIR)/logs\n" +\
        "\tmkdir -p $(RUN_DIR)/figures\n"
    lines.append(run_dir_job)
    initialize = "initialize: $(RUN_DIR)/catalog_samples.hdf5 build\n" +\
        '\techo "Successfully build gwalk and found run_dir"\n'
    lines.append(initialize)

    #### Generate Coord Tags ####
    # Check aligned3d
    if not COORD_TAG_RUNS[0] in coord_tags:
        print(fields)
        raise RuntimeError("Cannot produce aligned3d for %s"%group)

    # Identify prerequisites
    prereq = []
    for tag in coord_tags:
        # Skip bad tags
        if not tag in COORD_TAG_RUNS:
            continue
        # Initialize tag line
        tag_lines = []

        # Fit simple job
        line = r"%." + "%s_%s_fit_simple: "%(tag,group)
        if tag == COORD_TAG_RUNS[0]:
            line = line + "initialize\n"

        else:
            for item in prereq:
                line = line + "\\\n\t\t%." + "%s_%s_plot_select "%(item,group)
            line = line + '\n'
        tag_lines.append(line)
        tag_lines.append("\tpython3 %s %d fit-simple $(basename $@) %s %s\n"%(
                                            MAKE_GWALK, RUN_ID, group, tag))
        # Plot simple job
        tag_lines.append(
            "%." + "%s_%s_plot_simple: "%   (tag,group) +\
            "%." + "%s_%s_fit_simple\n"%    (tag,group)
                        )
        tag_lines.append("\tpython3 %s %d plot-simple $(basename $@) %s %s\n"%(
                                            MAKE_GWALK, RUN_ID, group, tag))
        # Fit genetic job
        tag_lines.append(
            "%." + "%s_%s_fit_genetic: "%   (tag,group) +\
            "%." + "%s_%s_plot_simple \n"%  (tag,group)
                        )
        tag_lines.append("\tpython3 %s %d fit-genetic $(basename $@) %s %s\n"%(
                                            MAKE_GWALK, RUN_ID, group, tag))
        # Plot genetic job
        tag_lines.append(
            "%." + "%s_%s_plot_genetic: "%  (tag,group) +\
            "%." + "%s_%s_fit_genetic \n"%  (tag,group)
                        )
        tag_lines.append("\tpython3 %s %d plot-genetic $(basename $@) %s %s\n"%(
                                            MAKE_GWALK, RUN_ID, group, tag))
        # Fit select job
        tag_lines.append(
            "%." + "%s_%s_fit_select: "%   (tag,group) +\
            "%." + "%s_%s_plot_genetic \n"%  (tag,group)
                        )
        tag_lines.append("\tpython3 %s %d fit-select $(basename $@) %s %s\n"%(
                                            MAKE_GWALK, RUN_ID, group, tag))
        # Plot genetic job
        tag_lines.append(
            "%." + "%s_%s_plot_select: "%  (tag,group) +\
            "%." + "%s_%s_fit_select \n"%  (tag,group)
                        )
        tag_lines.append("\tpython3 %s %d plot-select $(basename $@) %s %s\n"%(
                                            MAKE_GWALK, RUN_ID, group, tag))
        
        # Append tag lines
        for item in tag_lines:
            lines.append(item)
        tag_lines = []
        prereq.append(tag)
    
    # Define All
    lines.append("all: $(%s)\n"%group)
    # Define group
    lines.append("$(%s)"%group + ": %: %." + "%s_%s_plot_select\n"%(prereq[-1],group))

    with open(fname_group, 'w') as F:
        F.writelines(lines)




'''\
guess.py

Attempt to start random walkers in sensible locations

here be hacky bits that may or may not work

Beware!
'''
######## Imports ########
from os.path import join
import numpy as np
from .data import Database
from .coordinates import coord_tags as COORD_TAGS

######## Functions ########
def catalog_sample_guess(Norm, grid):
    '''\
    Load the catalog samples and take the mean and covariance of them
    '''
    posterior, prior = grid.load_catalog_samples()
    inv_prior = np.power(prior, -1.)
    X = Norm.fit_simple(posterior, w=inv_prior, assign=False).flatten()
    return X

def like_fits(Norm, wkdir, release, event, coord_tag):
    '''\
    find like fits, and use the results as a guess
    '''
    print("Finding like fits")
    fname_release = join(wkdir, "%s.nal.hdf5"%(release))
    db_release = Database(fname_release, event)
    labels = db_release.list_items()
    guesses = []
    for item in labels:
        if item.split(':')[0] == coord_tag:
            if db_release.exists(join(item,"mean")):
                mean = db_release.dset_value(join(item,"mean"))
                cov  = db_release.dset_value(join(item,"cov"))
                X = Norm.params_of_mean_cov(mean,cov).flatten()
                guesses.append(X)

    return guesses

def sort_guess(Norm, grid, guess_list):
    '''\
    find the best guess by kl divergence
    '''
    from numpy.linalg import eigvals
    guess_array = np.asarray(guess_list)
    nguess = guess_array.shape[0]
    guess_list = []
    kl = []
    hist_data = grid.load_catalog_hist()
    for i in range(nguess):
        try:
            item_kl = Norm.kl_div(
                                  grid.kl_samples,
                                  params=guess_array[i],
                                  hist_data = hist_data,
                                  coords = grid.coords,
                                  kl_bins = grid.kl_bins,
                                  kl_sensitivity = grid.kl_sensitivity,
                                  mode="component",
                                 )
            guess_list.append(guess_array[i])
            kl.append(item_kl)
            print("pass", eigvals(Norm.cov_of_params(guess_array[i])), guess_array[i])
        except Exception as exp:
            print("fail", eigvals(Norm.cov_of_params(guess_array[i])), guess_array[i])
            pass

    # Mix and match
    kl_tmp = np.asarray(kl)
    guess_mix = guess_list[0].copy()
    for j in range(len(guess_mix)):
        guess_mix[j] = guess_list[np.argmin(kl_tmp[:,j])][j]
    try:
        kl.append(Norm.kl_div(
                              grid.kl_samples,
                              params=guess_mix,
                              hist_data = hist_data,
                              coords = grid.coords,
                              kl_bins = grid.kl_bins,
                              kl_sensitivity = grid.kl_sensitivity,
                              mode="component",
                             ))
        guess_list.append(guess_mix)
    except:
        pass

    guess_array = np.asarray(guess_list)
    kl = np.asarray(kl)
    print("Guess list length: %d"%nguess)
    print("Guess kl: ")
    kl_sum = np.sum(kl[:,:grid.ndim], axis=1) + np.sum(kl[:,2*grid.ndim:],axis=1)
    kl_sum /= (grid.ndim * (grid.ndim + 1))//2
    print(kl_sum)
    index = np.argsort(kl_sum)
    guess_array = guess_array[index]
    kl_sum = kl_sum[index]
    return list(guess_array), kl_sum

def fill_guesses(Norm, grid, guess_list, nfill):
    '''\
    Find extra guesses based on existing guesses
    '''
    import time
    from scipy.stats import multivariate_normal
    from numpy.linalg import eigvals
    # Contain scope
    guess_list = np.asarray(guess_list).copy()
    print("top of guess list")
    print(guess_list[0])
    # Load catalog histogram data
    hist_data = grid.load_catalog_hist()
    # find number of guesses
    nguess = guess_list.shape[0]
    # Initialize new guesses
    new_guesses = np.zeros((nfill, guess_list.shape[1]))
    # Copy existing guesses
    new_guesses[:nguess] = guess_list
    # generate parameter mean and variance
    param_mean = np.average(guess_list, axis=0)
    #param_cov = np.cov(guess_list.T)
    param_cov = np.diag(0.01*np.ones(guess_list.shape[1]))
    rv = multivariate_normal(param_mean, param_cov, seed=Norm.rs)
    print("Dim: %d"%grid.ndim)
    print("Needed samples: %d"%(nfill - nguess))
    t_start = time.time()
    for i in range(nguess,nfill):
        i_complete = False
        while not i_complete:
            params = rv.rvs()
            try:
                kl = Norm.kl_div(
                                 grid.fit_samples,
                                 params=params,
                                 hist_data = hist_data,
                                 coords = grid.coords,
                                 kl_bins = grid.kl_bins,
                                 kl_sensitivity = grid.kl_sensitivity,
                                )
                new_guesses[i] = params
                i_complete = True
            except:
                pass
            
    t_end = time.time()
    print("Fill time: %f seconds"%(t_end - t_start))
    return new_guesses



def parameters_of_ndim(n):
    '''\
    Return the number of parameters associated witha guess in n dimensions
    '''
    return int((n*n + 3*n)//2)

def ndim_of_guess(guess, max_dim=20):
    '''\
    find the dimensionality of a guess
    '''
    for i in range(max_dim):
        if guess.size == parameters_of_ndim(i):
            return i
    raise RuntimeError("Failed to get number of dimensions for guess")

def coord_subset(coords1, coords2):
    '''\
    Check if coords1 is a subset of coords2
    '''
    truth = True
    for item in coords1:
        if not (item in coords2):
            truth = False
    return True

def visitor_NAL(grid, label):
    '''\
    This loads a fit from an existing NAL object
    '''
    from .NormalApproximateLikelihood import NAL
    coord_tag = label.split(":")[0]
    fit_method = label.split(":")[-1]
    Norm = NAL.load_fit(grid.fname_release,grid.group,grid.event,coord_tag,fit_method)
    return Norm


def event_fits(Norm, grid, best):
    '''\
    find like fits, and use the results as a guess
    '''
    print("Finding like fits")
    # Identify the coordinates
    coords = grid.coords
    # Protect the best guess
    best = np.copy(best)
    # Point at the database
    db_release = Database(grid.fname_release, grid.event)
    # List the fits for a given event
    labels = db_release.list_items()
    # Initialize list of guesses
    guesses = []
    # Begin fit loop
    for label in labels:
        # Check if fit exists
        try:
            Vis = visitor_NAL(grid, label)
        except: 
            Vis = None

        if not (Vis is None):
            # Load data
            # Identify the coordinates associated with a given fit
            coord_tag = label.split(":")[0]
            fit_coords = COORD_TAGS[coord_tag]
            # Program visitor


            ## Case 0: coordinates are identical ##
            if coord_tag == grid.coord_tag:
                # We already checked these models
                X = Vis.read_guess()

            ## Case 1: fit_coords is a subset of current coords ##
            else:
                # Map coordinates
                p_map = {}
                for i, item in enumerate(coords):
                    p_map[i] = None
                    for j, jtem in enumerate(fit_coords):
                        if item == jtem:
                            p_map[i] = j

                # Generate mean, var, and cor of best guess in Norm's scale
                mean = Norm.mean_of_params(best)[0]
                var = Norm.var_of_params(best)[0]
                cor = Norm.cor_of_params(best)[0]
                scale_norm = Norm.scale
                # Generate mean, var, and cor of Vis in Vis's scale
                mean_vis = Vis.guess_mean()
                var_vis = Vis.guess_var()
                cor_vis = Vis.guess_cor()
                scale_vis = Vis.scale
                # Loop through each coordinate in Norm's coordinate system
                for i in range(grid.ndim):
                    # Do nothing if coordinate isn't in Vis' coordinate system
                    if not (p_map[i] is None):
                        # Find the scale
                        scale_i = scale_vis[p_map[i]]/scale_norm[i]
                        # Update the mean
                        mean[i] = mean_vis[p_map[i]] * scale_i
                        var[i] = var_vis[p_map[i]] * scale_i
                        for j in range(i):
                            if not (p_map[j]) is None:
                                cor[i,j] = cor[j,i] = cor_vis[p_map[i],p_map[j]]
                # Get params back
                X = []
                for i in range(grid.ndim):
                    X.append(mean[i])
                for i in range(grid.ndim):
                    X.append(var[i])
                for i in range(grid.ndim):
                    for j in range(i):
                        X.append(cor[i,j])
                X = np.asarray(X)

            guesses.append(X)

    return guesses

def event_fit_modification(Norm, best, fits):
    '''\
    Make modification to event fits to use inferred correlation
    '''
    # Contain scope
    best = np.asarray(best).flatten().copy()
    ndim = Norm.ndim
    fits = np.asarray(fits).copy()
    # Initialize new fits
    new_fits = []
    for i in range(fits.shape[0]):
        item = fits[i].copy()
        item[2*ndim:] = best[2*ndim:]
        new_fits.append(item)

    return new_fits

def hist1d_guess(Norm, grid, best):
    '''\
    Make guesses based on the quadratic nature of the log likelihood
    '''
    print("\nHistogram guess!")
    best = best.copy()
    hist_data = grid.load_catalog_hist()
    edges = hist_data["edges"]
    centers = hist_data["centers"]
    X = np.copy(best)
    for i in range(grid.ndim):
        hist1d = hist_data["hist1d_%s"%grid.coords[i]]
        #X[i] = centers[i,np.argmax(hist1d)]
        hist1d = hist1d / np.sum(hist1d)
        keep = hist1d > 0
        x = centers[i][keep] / Norm.scale[i]
        y = np.log(hist1d[keep])
        a, b, c = np.polyfit(x, y, 2)
        if a < 0:
            X[i] = (-0.5*b/a )# / Norm.scale[i]
            X[i + grid.ndim] = (-0.5/a)# / Norm.scale[i]
        if not Norm.satisfies_constraints(X):
            X[i] = centers[i,np.argmax(hist1d)] / Norm.scale[i]
            X[i + grid.ndim] = best[i + grid.ndim]
        if not Norm.satisfies_constraints(X):
            X[i] = best[i]
            X[i + grid.ndim] = best[i + grid.ndim]
    # Rescale by standard deviation
    #X[:Norm.ndim] /= Norm.scale
    #X[Norm.ndim:2*Norm.ndim] /= Norm.scale
        
    return X

######## Primary Algorithm ########
def make_guess(Norm, grid, nfill = None):
    '''\
    Make educated guesses on locations to start the random walkers in.

    Beware: This is realy hacky!

    Inputs: Grid object

    Outputs: X_g - param guesses
    '''
    print("Loaded guess algorithm")
    #### Gather Information ####
    ## Simple attrs ##
    cata = grid.cata
    wkdir = grid.wkdir
    event = grid.event
    group = grid.group
    coord_tag = grid.coord_tag
    fname_grid = grid.fname
    fname_cata = grid.fname_cata
    release = grid.release
    ndim = grid.ndim
    coords = grid.coords
    limits = grid.limits
    fit_samples = grid.fit_samples
    kl_bins = grid.kl_bins
    kl_samples = grid.kl_samples
    kl_sensitivity = grid.kl_sensitivity

    #### Initialize ####
    guess_list = []

    #### Guesswork ####
    ## Primary Guesses ##
    # First guess: Whatever the Norm object was initialized with
    guess_list.append(Norm.read_guess())
    # Second guess: Mean and covariance of catalog samples
    guess_list.append(catalog_sample_guess(Norm, grid))
    best = guess_list[1].copy()

    ## Secondary Guesses ##
    # Histogram inspired guess
    guess_list.append(hist1d_guess(Norm, grid, best))

    ## Tertiary Guesses ##
    # These come from real fits to the event, 
    #   but may have different parameters
    like_fits = event_fits(Norm, grid, best)
    for item in like_fits:
        guess_list.append(item.copy())
    # Generate additional fits based on real fits with correlation parameters
    # matching the simple fit
    mod_fits = event_fit_modification(Norm, best, like_fits)
    for item in mod_fits:
        guess_list.append(item.copy())

    ## Evaluation of guesses ##
    print("Sorting guess list...")
    guess_list, kl_sum = sort_guess(Norm, grid, guess_list)

    ## Extra guesses ##
    # Generate extra guesses to fill number requirement
    if (not (nfill is None)) and (len(guess_list) < nfill):
        new_guesses = fill_guesses(Norm, grid, guess_list, nfill)
        guess_list = new_guesses
   
    return np.asarray(guess_list), kl_sum



#!/usr/bin/env python3
'''\
The Multivariate Normal object for fitting

Provide infrastructure for models and hyperparameters
'''
#### Module imports ####
from . import Parameter, Relation, Covariance
from ..utils.multivariate_normal import cov_of_std_cor
from ..utils.multivariate_normal import std_of_cov
from ..utils.multivariate_normal import cor_of_cov
from ..utils.multivariate_normal import mu_of_params
from ..utils.multivariate_normal import std_of_params
from ..utils.multivariate_normal import cor_of_params
from ..utils.multivariate_normal import cov_of_params
from ..utils.multivariate_normal import mu_cov_of_params
from ..utils.multivariate_normal import params_of_mu_cov
from ..utils.multivariate_normal import multivariate_normal_pdf
from ..utils.multivariate_normal import n_param_of_ndim
from ..utils.multivariate_normal import ndim_of_n_param
from ..utils.multivariate_normal import params_reduce_1d
from ..utils.multivariate_normal import params_reduce_2d
from ..utils.multivariate_normal import params_reduce_dd
from ..utils.multivariate_normal import multivariate_normal_marginal1d
from ..utils.multivariate_normal import multivariate_normal_marginal2d
from relative_entropy_cython import relative_entropy
from ..utils.convergence import density_to_gaussian_kl

#### Public Imports ####
import numpy as np
#### Globals ####
_INV_RT2 = 1./np.sqrt(2)
_LOG_2PI = np.log(2*np.pi)

######## Multivariable Normal Object ########
class MultivariateNormal(Relation):
    '''\
    fit data to a Gaussian and provide different methods
    '''
    #### Initialize ####
    def __init__(
                 self,
                 variables,
                 scale,
                 random_state=None,
                 sig_max =3.,
                ):
        '''\
        Initialize a Multivariate Normal Relation

        Inputs: variables - A list, tuple, or dictionary of parameter objects
                            This does not include the covariance matrix
                            which we will generate automatically
                            to reduce pain
                            Variables specifically describe the domain
                                of the function we are fitting.

        '''
        # Check random state
        if random_state is None:
            self.rs = np.random.RandomState()
        else:
            self.rs = random_state

        # Identify the number of dimensions in the domain
        self.ndim = len(variables)

        # Generate the covariance parameters
        cov_parameters = Covariance(
                                          self.ndim,
                                          sig_max=sig_max,
                                          random_state=random_state,
                                         )
        # Keep track of the scale
        self.scale = scale*np.ones(self.ndim)

        # Generate a temporary relationship with the variables
        temp = Relation(variables,random_state=self.rs)

        # Inialize a Relation object with the mu variables
        super().__init__(random_state=self.rs)
        mu_parameters = Relation(random_state=self.rs)

        # Rename mu variables
        for i, key in enumerate(temp._parameters):
            p = Parameter(
                          "mu_%d"%(i),
                          temp._parameters[key].guess/self.scale[i],
                          temp._parameters[key].limits/self.scale[i],
                          label=temp._parameters[key].label
                         )
            mu_parameters.add_parameter(p)

        # Contain the entire list of covariance parameters
        temp_obj = mu_parameters + cov_parameters
        self.__dict__.update(temp_obj.__dict__)

        # Carry functions that identify things about the covariance matrix
        self.read_std      = cov_parameters.read_std
        self.read_cor      = cov_parameters.read_cor
        self.read_cov      = cov_parameters.read_cov
        self.params_of_cov  = cov_parameters.params_of_cov
        self.mu_hypercube  = mu_parameters.hypercube
        self.sample_uniform_variable = mu_parameters.sample_uniform


    #### Tools for working with this object ####
    ## Guess reading tools ##

    def read_mu(self):
        '''\
        Just read the mu guesses
        '''
        ## Imports ##
        # Public 
        import numpy as np
        # Initialize mu
        mu = np.zeros(self.ndim)
        # Loop through std parameters
        for i in range(self.ndim):
            # Generate tag
            tag = "mu_%d"%(i)
            # Load value
            mu[i] = self._parameters[tag].guess
        return mu

    def read_scaled(self):
        '''\
        Read the guess to mu and cov
        '''
        mu = self.read_mu()
        cov = self.read_cov()
        return mu, cov

    def read_physical(self):
        '''\
        Read the guess to mu and cov
        '''
        mu = self.read_mu()*self.scale
        cov = self.read_cov()*np.outer(self.scale,self.scale)
        return mu, cov

    ## Conversions ##

    def mu_of_params(self,X):
        '''\
        Convert parameters to mu values
        '''
        return mu_of_params(X)

    def std_of_params(self,X):
        '''\
        Convert parameters to std values
        '''
        return std_of_params(X)

    def cor_of_params(self, X):
        '''\
        Convert parameters to corelation values
        '''
        return cor_of_params(X)

    def cov_of_params(self, X):
        '''\
        Convert parameters to covariance values
        '''
        return cov_of_params(X)

    def mu_cov_of_params(X):
        '''\
        Convert parameter values to mu and cov values
        '''
        return mu_cov_of_params(X)

    def params_of_mu_cov(mu,cov):
        '''\
        Convert parameter values to mu and cov values
        '''
        return params_of_mu_cov(mu,cov)

    ## Guess assignment ##

    def satisfies_constraints(self, Xs):
        '''\
        Pick only valid guesses for params
        '''
        ## Imports ##
        # Public
        import numpy as np
        from scipy.stats import multivariate_normal
        # Fix data
        Xs = self.check_sample(Xs)
        # Check mu parameters
        c1 = self.summands[0].satisfies_constraints(Xs[:,:self.ndim])
        # Check covariance
        c2 = self.summands[1].satisfies_constraints(Xs[:,self.ndim:])
        constrained_indx = (c1 & c2)
        # Print statement for errors
        #if (constrained_indx.size == 1) and (not constrained_indx):
        #    print(c1, c2, c3)

        return constrained_indx
                
    def likelihood(
                   self,
                   Y,
                   X=None,
                   w=None,
                   return_prod=False,
                   log_scale=False,
                   scale=False,
                   indices=None,
                  ):
        '''\
        Find the likelihood of some data with a particular guess of parameters
        
        Inputs:
            X - parameter samples (mu_x, mu_y, mu_z, std_x, ...)
            Y - Variable samples (r_x, r_y, ...)
            w - weights: You would only use these if taking a sum
            scale - describes how to scale input data
                if False: assume input data is PHYSICAL
                if True: assume input data is SCALED
                if (len(Y),) array: scale input by given values
                if (ngauss, len(Y)): Each sample gaussian has its own scale

        Outputs:
            L - array of likelihoods ((len(X), len(Y))
        '''
        ## Imports ##
        # Public
        import numpy as np
        from scipy.stats import multivariate_normal
        from scipy import stats

        # Protect scope
        Y = Y.copy()

        ## Check X ##
        # If X is None, guess is already assigned
        if X is None:
            X = self.read_guess()
        # Check the sample
        X = self.check_sample(X)

        # Read information
        n_gauss, n_param = X.shape
        ndim_X = ndim_of_n_param(n_param)
        assert ndim_X == self.ndim

        # Check indices
        if indices is None:
            indices = np.arange(ndim_X)
        else:
            indices = np.asarray(indices)
        ndim_Y = indices.size

        ## Check Y ##
        # Check is numpy array
        if not isinstance(Y, np.ndarray):
            Y = np.asarray(Y)
        n_sample = Y.shape[0]

        # Check dimensions
        Y = Y.reshape((n_sample, ndim_Y))
            
        ## Check scale ##
        if scale is False:
            scale = np.ones((n_gauss,ndim_X))*self.scale
        elif scale is True:
            scale = np.ones((n_gauss,ndim_X))
        elif np.asarray(scale).size == ndim_X:
            scale = np.ones((n_gauss,ndim_X))*scale
        elif np.asarray(scale).shape == (ngauss,ndim_X):
            # This is intended behavior
            pass
        else:
            raise RuntimeError("Strange Scale. Aborting.")

        # Isolate indices
        X = params_reduce_dd(X, indices)
        scale = scale[:,indices]
      
        # Evaluate normal pdf
        L = multivariate_normal_pdf(X, Y, scale=scale)

        # Apply weights
        if not (w is None):
            if X is None:
                L = np.atleast_2d(L.dot(w))
            elif (len(X.shape) == 2) and (X.shape[1] == len(self._parameters)):
                L *= w 
                
        # Check how we want the output
        if log_scale and return_prod:
            L = np.sum(np.log(L), axis=1)
        elif log_scale:
            L = np.log(L)
        elif return_prod:
            L = np.prod(L, axis=1)
        else:
            L = L

        return L

    def pdf(self,Y):
        '''\
        Return the pdf using the current guess for the likelihood
        '''
        return self.likelihood(Y)

    def lnL(self,Y,w=None):
        '''\
        Return the sum of the log of the likelihood
        '''
        return self.likelihood(Y,log_scale=True,return_prod=True,w=w)

    #### Resample ####

    def sample_normal_unconstrained(self, rv, size):
        '''\
        Draw a number, size, of random samples
        '''
        ## Imports ##
        # Public
        import numpy as np
        # Generate samples
        samples = np.atleast_2d(rv.rvs(size = int(size),random_state=self.rs))
        if not samples.shape[1] == self.ndim:
            raise RuntimeError("normal samples are the wrong shape")
        return samples

    def sample_normal(self, size, scale=False, params=None):
        '''\
        Draw a number, size, of random samples from inside limits

        This introduces bias, so be careful
        '''
        ## Imports ##
        # Public
        import numpy as np
        from scipy.stats import multivariate_normal

        # Initialize samples
        sample_shape = (size, self.ndim)
        Xk = np.zeros((size, self.ndim))

        # Initialize the number of samples we have
        n_keep = 0

        # load params
        if params is None:
            mu = self.read_mu()
            cov = self.read_cov()
        else:
            mu, cov = mu_cov_of_params(params)
            mu = mu[0]
            cov = cov[0]

        # Initialize rv
        try:
            rv = multivariate_normal(mu, cov)
        except RuntimeError as exp:
            from numpy.linalg import eigvals, eigvalsh
            print("Sampling failed")
            print("Satisfies constraints: ")
            print(self.satisfies_constraints(params))
            print("scale: ")
            print(self.scale)
            print("std_dev: ")
            print(std_of_cov(cov))
            print("cor: ")
            print(cor_of_cov(cov))
            print("eig: ")
            print(eigvalsh(cov))
            raise exp

        # Continue doing this until we have the right number of samples
        while n_keep < size:
            # How many samples do we still need?
            n_need = size - n_keep
            # Draw that many samples
            Xs = self.sample_normal_unconstrained(rv, n_need)
            # Check if these samples satisfy our constraints
            k_indx = self.summands[0].satisfies_constraints(Xs)
            # Identify the number of new samples to keep from this iteration
            n_it = np.sum(k_indx)
            # Keep iterating if there are not valid samples
            if n_it != 0:
                # Keep valid samples
                Xs = Xs[k_indx]
                # Add them to Xkeep
                Xk[n_keep:n_keep+n_it,:] = Xs
                # Update n_keep
                n_keep += n_it

        # Rescale if applicable
        if not scale:
            Xk *= self.scale

        # Return samples
        return Xk

    #### Basic Fit Method ####

    def fit_simple(self, Y, w=None, assign=True):
        '''\
        Find the mean and covariance of random samples

        Inputs:
            - Y: random samples (n_samples, n_v)
                where nv is the number of variables, not parameters
            - weights: weights for weighted prior

        Return:
            - mean: the mean of the multivariate normal distribution (nd,)
            - cov: The covariance matrix of the multivariate
                normal distribution (nd, nd)
        '''
        ## Imports ##
        import numpy as np

        # Find the average and covariance of the samples
        Y = Y/self.scale
        mean = np.average(Y, weights = w, axis = 0)
        cov = np.cov(Y.T, aweights = w)

        # convert to a sample
        X = params_of_mu_cov(mean, cov)

        # Assign this guess
        if assign:
            self.assign_guess(X)

        # Return the likelihood
        return X

    #### Fit mesh ####

    def fit_mesh(
                 self,
                 mesh,
                ):
        '''\
        Fit the mesh using different methods
        '''
        # Identify number of guesses
        nguess = 3
        # Initialize guesses
        Xg = np.tile(self.read_guess(),(nguess,1))

        # Generate 1D evaluation guesses
        for i in range(self.ndim):
            # Load mesh values
            y_test, L_test = mesh.fetch_1d_evaluations(i)
            # adjust coordinates
            keep = L_test > 0
            y_test = y_test[keep].flatten()/self.scale[i]
            L_test = np.log(L_test[keep].flatten())
            # Find maximum
            a, b, c = np.polyfit(y_test, L_test, 2)
            if a < 0:
                Xg[0,i] = (-0.5*b/a)
                Xg[0,i+self.ndim] = (-0.5/a)
            else:
                # Toss end points
                _test = y_test[1:-1].flatten()
                L_test = L_test[1:-1].flatten()
                Xg[0,i] = y_test[np.argmax(L_test)]

        # Generate 1D training guesses
        for i in range(self.ndim):
            # Load data
            y_train = mesh.marginals["1d_%d_x_train"%i]
            L_train = mesh.marginals["1d_%d_y_train"%i]
            bins = int(mesh.marginals["1d_%d_bins"%i])
            # Rescale training set 1
            y_train_1 = y_train[:bins].flatten()/self.scale[i]
            L_train_1 = L_train[:bins].flatten()
            keep_1 = L_train_1 > 0
            y_train_1 = y_train_1[keep_1]
            L_train_1 = np.log(L_train_1[keep_1])

            # Training set 1 fit
            a, b, c = np.polyfit(y_train_1, L_train_1, 2)
            if a < 0:
                Xg[1,i] = (-0.5*b/a)
                Xg[1,i+self.ndim] = (-0.5/a)
            else:
                # Toss end points
                y_train_1 = y_train_1[1:-1].flatten()
                L_train_1 = L_train_1[1:-1].flatten()
                Xg[1,i] = y_train_1[np.argmax(L_train_1)]

            # Rescale training set 2
            y_train_2 = y_train[bins:].flatten()/self.scale[i]
            L_train_2 = L_train[bins:].flatten()
            keep_2 = L_train_2 > 0
            y_train_2 = y_train_2[keep_2]
            L_train_2 = np.log(L_train_2[keep_2])

            # Training set 1 fit
            a, b, c = np.polyfit(y_train_2, L_train_2, 2)
            if a < 0:
                Xg[2,i] = (-0.5*b/a)
                Xg[2,i+self.ndim] = (-0.5/a)
            else:
                # Toss end points
                y_train_2 = y_train_2[1:-1].flatten()
                L_train_2 = L_train_2[1:-1].flatten()
                Xg[2,i] = y_train_2[np.argmax(L_train_2)]


        return Xg

    #### Convergence ####

    def pdf_kl(
               self,
               X,
               Y_pdf=None,
               L_pdf=None,
               eps=0.,
               kl_sensitivity=None,
               **kwargs
              ):
        '''\
        Given some points, fit the marginal kl divergences
        '''
        # Make sure the function was called correctly
        assert not(Y_pdf is None)
        assert not(L_pdf is None)
        # Consider only valid guesses for X
        X = self.check_sample(X)
        # Identify number of gaussians
        n_gauss = X.shape[0]

        # Generate likelihood values for each model
        L_X = self.likelihood(Y_pdf, X=X)
        L_X /= np.sum(L_X,axis=-1)[:,None]

        # Incorperate kl sensitivity
        if not (kl_sensitivity is None):
            kl_sensitivity = kl_sensitivity*np.max(L_pdf)
            L_X =   ((1. - kl_sensitivity)*L_X)     + kl_sensitivity
            L_pdf = ((1. - kl_sensitivity)*L_pdf)   + kl_sensitivity

        kl = relative_entropy(L_pdf, L_X)

        return kl

    def mesh_kl(
                self,
                density_mesh,
                X,
                mode = "mean",
                kl_sensitivity=None,
                **kwargs
               ):
        '''\
        Evaluate the 1D and 2D kl divergences between a mesh and normal model
        '''
        # Consider only valid guesses for X
        X = self.check_sample(X)

        ## Calculate kl divergences ##
        kl1d, kl2d  = \
            density_to_gaussian_kl(
                                   density_mesh,
                                   X,
                                   scale=self.scale,
                                   kl_sensitivity=kl_sensitivity,
                                  )

        # Calculate the goodness of fit statistic
        if mode == "sum":
            kl = np.sum(kl1d,axis=-1) + np.sum(np.sum(kl2d,axis=-1),axis=-1)
        elif mode == "rms":
            kl = np.sqrt(np.sum(kl1d**2,axis=-1) + np.sum(np.sum(kl2d**2,axis=-1),axis=-1))
        elif mode == "mean":
            kl1d_sum = np.sum(kl1d,axis=-1)
            kl2d_sum = np.sum(np.sum(kl2d,axis=-1),axis=-1)
            kl = 0.5 * (
                        (kl1d_sum/self.ndim) +\
                        (kl2d_sum/(self.ndim*(self.ndim -1)//2))
                       )
            kl_ind = np.argmin(kl)
        elif mode == "component":
            kl = np.empty((X.shape[0],len(self._parameters)-self.ndim))
            # First consider the 1D kl divergences
            for i in range(self.ndim):
                kl[:,i] = kl1d[:,i]
            # The 2d kl divergences have a bearing on 5 different parameters
            for i in range(self.ndim):
                for j in range(i):
                    kl[:,self.p_map["cor_%d_%d"%(i,j)]-self.ndim] = kl2d[:,i,j]
        elif mode == "parameter":
            kl = np.zeros((X.shape[0],len(self._parameters)))
            # First consider the 1D kl divergences
            for i in range(self.ndim):
                kl[:,i] = kl1d[:,i]
                kl[:,i+self.ndim] = kl1d[:,i]
            # The 2d kl divergences have a bearing on 5 different parameters
            for i in range(self.ndim):
                for j in range(i):
                    kl[:,i] += kl2d[:,i,j]
                    kl[:,i+self.ndim] += kl2d[:,i,j]
                    kl[:,j] += kl2d[:,i,j]
                    kl[:,j+self.ndim] += kl2d[:,i,j]
                    kl[:,self.p_map["cor_%d_%d"%(i,j)]] = kl2d[:,i,j]
            # for each parameter, we want to use the average of the
            # kl divergences associated with that parameter.
            kl[:,:self.ndim] /= 3
            kl[:,self.ndim:2*self.ndim] /= 3
        else:
            raise RuntimeError("Unknown kl mode: %s"%(mode))

        return kl

    def mesh_kl_alt(
                self,
                density_mesh,
                X,
                size=1000000,
                mode = "mean",
                eps=0.,
                kl_sensitivity=None,
               ):
        '''\
        Evaluate the 1D and 2D kl divergences between a mesh and normal model
        '''
        from scipy.special import rel_entr, kl_div
        from fast_histogram import histogram1d, histogram2d
        # Consider only valid guesses for X
        X = self.check_sample(X)

        # We will only use the first ndim limits
        limits = self.read_limits()[:self.ndim]

        # Generate normal samples
        Y_norm = self.sample_normal(size, params=X, scale=True)

        ## Calculate 1D kl divergences ##
        # Initialize kl divergence
        kl1d = np.zeros((1,self.ndim,))
        for i in range(self.ndim):
                # Histogram each dimension
                H_uni = density_mesh.marginals["1d_%d_y_test"%i]
                H_cur = histogram1d(Y_norm[:,i],range=limits[i],bins=H_uni.size)

                # Renormalize
                H_uni = H_uni/np.sum(H_uni)
                H_cur = H_cur/np.sum(H_cur)

                # Generate a 1d kl divergence
                # Protect against zeros
                H_uni = (1. - kl_sensitivity)*H_uni + kl_sensitivity
                H_cur = (1. - kl_sensitivity)*H_cur + kl_sensitivity

                # calculate the sum of the relative entropy in a direction
                kl1d[:,i] = np.sum(rel_entr(H_uni, H_cur))

        ## Calculate 2D kl divergences ##
        # The ones we don't use will just stay zero for the sum
        kl2d = np.zeros((1,self.ndim, self.ndim))

        for i in range(self.ndim):
            for j in range(i):
                    H_uni = density_mesh.marginals["2d_%d_%d_y_test"%(i,j)]
                    # Histogram each pair
                    H_cur = histogram2d(
                                        Y_norm[:,i], Y_norm[:,j],
                                        range=[limits[i],limits[j]],
                                        bins=int(np.sqrt(H_uni.size)),
                                       ).flatten()

                    # Renormalize
                    H_uni = H_uni/np.sum(H_uni)
                    H_cur = H_cur/np.sum(H_cur)

                    # Generate a 1d kl divergence
                    # Protect against zeros
                    H_uni = (1. - kl_sensitivity)*H_uni + kl_sensitivity
                    H_cur = (1. - kl_sensitivity)*H_cur + kl_sensitivity

                    # calculate the sum of the relative entropy in a direction
                    kl2d[:,i,j] = np.sum(rel_entr(H_uni, H_cur))


        # Calculate the goodness of fit statistic
        if mode == "sum":
            kl = np.sum(kl1d,axis=-1) + np.sum(np.sum(kl2d,axis=-1),axis=-1)
        elif mode == "rms":
            kl = np.sqrt(np.sum(kl1d**2,axis=-1) + np.sum(np.sum(kl2d**2,axis=-1),axis=-1))
        elif mode == "mean":
            kl = 0.5*((np.sum(kl1d,axis=-1)/len(kl1d)) + \
                (np.sum(np.sum(kl2d,axis=-1),axis=-1) / \
                np.sum(np.sum(kl2d != 0.,axis=-1),axis=-1)))
        elif mode == "component":
            kl = np.empty((X.shape[0],len(self._parameters)-self.ndim))
            # First consider the 1D kl divergences
            for i in range(self.ndim):
                kl[:,i] = kl1d[:,i]
            # The 2d kl divergences have a bearing on 5 different parameters
            for i in range(self.ndim):
                for j in range(i):
                    kl[:,self.p_map["cor_%d_%d"%(i,j)]-self.ndim] = kl2d[:,i,j]
        elif mode == "parameter":
            kl = np.zeros((X.shape[0],len(self._parameters)))
            # First consider the 1D kl divergences
            for i in range(self.ndim):
                kl[:,i] = kl1d[:,i]
                kl[:,i+self.ndim] = kl1d[:,i]
            # The 2d kl divergences have a bearing on 5 different parameters
            for i in range(self.ndim):
                for j in range(i):
                    kl[:,i] += kl2d[:,i,j]
                    kl[:,i+self.ndim] += kl2d[:,i,j]
                    kl[:,j] += kl2d[:,i,j]
                    kl[:,j+self.ndim] += kl2d[:,i,j]
                    kl[:,self.p_map["cor_%d_%d"%(i,j)]] = kl2d[:,i,j]
            # for each parameter, we want to use the average of the
            # kl divergences associated with that parameter.
            kl[:,:self.ndim] /= 3
            kl[:,self.ndim:2*self.ndim] /= 3
        else:
            raise RuntimeError("Unknown kl mode: %s"%(mode))

        return kl

            

    #### methods for random walkers ####

    def walker_init(
                    self,
                    convergence="mesh_kl",
                    nwalk=100,
                    mesh=None,
                    Y_pdf=None,
                    L_pdf=None,
                    **kwargs
                   ):
        #import concurrent.futures
        #from concurrent.futures import ThreadPoolExecutor
        #from time import time

        ## Generate initial placements for MCMC walkers in parameter space ##
        cur_0 = self.sample_uniform_unconstrained(nwalk)

        ## Handle Guesswork ##
        if ("guess_list" in kwargs) and not (kwargs["guess_list"] is None):
            guess_list = kwargs["guess_list"]
            nguess = guess_list.shape[0]
            # Rescale guess list
            if nguess > nwalk:
                nguess = nwalk
                guess_list = guess_list[:nwalk]
            # Apply guesses
            cur_0[:nguess] = guess_list

        else: 
            nguess = 0

        ## Make scattered placements ##

        cur_0[nguess:] = self.sample_uniform_unconstrained(nwalk - nguess)

        if convergence == "mesh_kl":
            # Assert mesh is not none
            assert not (mesh is None)
     

            def kl_likelihood(cur, **kwargs):
                # Prepare cur
                cur = np.atleast_2d(cur)
                # Initialize L
                L = np.zeros((cur.shape[0],))
                # Find nans
                keep = self.satisfies_constraints(cur)
                L[~keep] = 0.
                # Calculate likelihood
                if np.sum(keep) > 0:
                    kl = self.mesh_kl(
                                      mesh,
                                      cur[keep],
                                      **kwargs
                                     )
                      
                    L[keep] = np.power(kl,-1.)

                return L

        elif convergence == "mesh_kl_alt":
            raise NotImplemented

        elif convergence == "pdf_kl":
            # Assert things
            assert not (Y_pdf is None)
            assert not (L_pdf is None)
            # renormalize likelihood
            L_pdf /= np.sum(L_pdf)
            
            def kl_likelihood(cur, **kwargs):
                # Prepare cur
                cur = np.atleast_2d(cur)
                # Initialize L
                L = np.zeros((cur.shape[0],))
                # Find nans
                keep = self.satisfies_constraints(cur)
                L[~keep] = 0.
                # Calculate likelhood
                if np.sum(keep) > 0:
                    kl = self.pdf_kl(
                                     cur[keep],
                                     **kwargs
                                    )
                    L[keep] = np.power(kl, -1)
                return L

        return cur_0, kl_likelihood


    #### Emcee ####
    def fit_emcee(
                  self,
                  Y, P,
                  nwalk=100,
                  nstep=1000,
                  **kwargs
                 ):
        '''\
        Find the mu and covariance of random samples

        Inputs:
            - Y: random samples (n_samples, n_v)
                where nv is the number of variables, not parameters
            - weights: weights for weighted prior

        Return:
            - mu: the mu of the multivariate normal distribution (nd,)
            - cov: The covariance matrix of the multivariate
                normal distribution (nd, nd)
        '''
        ## Imports ##
        import numpy as np
        import emcee

        # Generate fit
        cur_0, f = self.walker_init(Y, P, **kwargs)
        sampler = emcee.EnsembleSampler(nwalk,cur_0.shape[1],f,vectorize=True)
        emcee_state = sampler.run_mcmc(cur_0, nstep)
        best_ind = np.argmax(emcee_state.log_prob)
        best_fit = emcee_state.coords[best_ind]
        self.assign_guess(best_fit)

        chains = sampler.get_chain()
        MLE_likelihood = sampler.get_log_prob().T
        MLE_params = np.transpose(chains, axes=[1,0,2])

        # Reshape output
        MLE_params = MLE_params.reshape((nwalk * nstep, len(self._parameters)))
        MLE_likelihood_flat = MLE_likelihood.reshape((nwalk * nstep,))

        # Find the maximal value
        imax = np.argmax(MLE_likelihood_flat)
        best_guess = MLE_params[imax,:]

        # Assign the best guess!
        self.assign_guess(best_guess)

        #return kl, MLE_likelihood
        
    #### Random Walk Step Functions ####

    def RW_uniform_step(self, L_fn, cur, **kwargs):
        '''\
        Draw a new step randomly within bounds, compare the likelihood

        Inputs:
            cur: curent walker samples
        '''
        # Check the number of walkers
        nwalk = kwargs["nwalk"]

        # Generate a new guess which satisfies constraints
        new = self.sample_uniform(nwalk, rs=rs)

        # Determine the likelihood of the new guess
        L_new = L_fn


        return new, L_new

    def RW_normal_step(
                       self,
                       L_fn,
                       cur,
                       scale=0.1,
                       carryover=0.01,
                       **kwargs
                      ):
        '''\
        Draw a new step randomly within bounds, compare the likelihood
        '''
        ## Imports ##
        # Public
        from scipy.stats import multivariate_normal

        # Check the number of walkers
        nwalk = kwargs["nwalk"]
        # Read limits
        limits = self.read_limits()
        # Protect from annealing too far
        sig = scale*(limits[:,1] - limits[:,0])
        # Initialize the number of valid samples we have
        n_keep = 0

        # Initialize new sample shape
        sample_shape = (nwalk, len(self._parameters))

        # Initialize kl array
        L_cur = L_fn(cur, **kwargs)

        # Identify best guesses
        n_carry = int(carryover*nwalk)
        if n_carry > np.sum(L_cur > 0):
            n_carry = np.sum(L_cur > 0)
        carry_index = np.argsort(L_cur)[-n_carry:]
        carry = cur[carry_index]
        L_carry = L_cur[carry_index]

        ## Generate new steps ##
        # Initialize new samples to keep
        new = np.zeros(sample_shape)
        # loop through each random walker
        for i, item in enumerate(cur):
            # Find a valid future step
            new[i] = multivariate_normal.rvs(item,sig,random_state=self.rs)
        keep = self.satisfies_constraints(new)
        new[~keep] = cur[~keep]

        # Determine the likelihood of the new guess
        L_new = L_fn(new, **kwargs)

        # Determine alpha
        alpha = L_new/L_cur

        # Decide if to jump
        jumpseed = (self.rs.uniform(size=nwalk) > (1 - alpha)).astype(bool)
        jumpseed[L_cur==0] = True

        # Jump
        new[~jumpseed] = cur[~jumpseed]
        L_new[~jumpseed] = L_cur[~jumpseed]

        # Hold best guesses
        drop_index = np.argsort(L_new)[:n_carry]
        new[drop_index] = carry
        L_new[drop_index] = L_carry

        return new, L_new

    def RW_genetic_step(
                        self,
                        L_fn,
                        cur,
                        carryover=0.03,
                        **kwargs
                       ):
        '''\
        Draw a new step randomly within bounds, compare the likelihood
        '''
        ## Imports ##
        # Public
        from scipy.stats import multivariate_normal

        # Check the number of walkers
        nwalk = kwargs["nwalk"]
        # Read limits
        limits = self.read_limits()
        # Use variance of guesses to determine jump scale
        sig = np.sqrt(np.var(cur, axis=0))
        # Initialize the number of valid samples we have
        n_keep = 0

        # Initialize new sample shape
        sample_shape = (nwalk, len(self._parameters))

        # Generate component kl divergences
        if kwargs["convergence"] == "mesh_kl":
            # Identify candidates to potentially keep
            keep = self.satisfies_constraints(cur)
            # Initialize kl array
            kl_array = np.zeros_like(cur)
            L_cur = np.zeros(nwalk)
            kl_array[keep] = self.mesh_kl(
                                          kwargs["mesh"],
                                          cur[keep],
                                          mode="parameter",
                                          **kwargs
                                         )
            L_cur[keep] = np.power(np.sum(kl_array[keep],axis=-1),-1.)

            # For all of the failed guesses, we need to set their likelihood
            L_cur[~keep] = 0.

        else:
            # Generate initial likelihood
            L_cur = L_fn(cur, **kwargs)

        # Identify best guesses
        n_carry = int(carryover*nwalk)
        if n_carry > np.sum(L_cur > 0):
            n_carry = np.sum(L_cur > 0)
        carry_index = np.argsort(L_cur)[-n_carry:]
        carry = cur[carry_index]
        L_carry = L_cur[carry_index]

        ## Breeding ##
        # The breeding pool excludes candidates with fitness zero
        keep = L_cur > 0
        Xb = cur[keep].copy()
        if kwargs["convergence"] == "mesh_kl":
            kl_b = kl_array[keep].copy()
        Lb = L_cur[keep].copy()
        Lb /= np.sum(Lb)
        # Pick random parents
        p1 = self.rs.choice(np.arange(Xb.shape[0]),size=nwalk,p=Lb)
        p2 = self.rs.choice(np.arange(Xb.shape[0]),size=nwalk,p=Lb)
        if kwargs["convergence"] == "mesh_kl":
            for i in range(nwalk):
                # Choose random parameter values
                choices = np.empty(cur.shape[1],dtype=bool)
                kl_b_sum = kl_b[p1[i]] + kl_b[p2[i]]
                for j in range(cur.shape[1]):
                    choices[j] = self.rs.choice(
                           [True,False],
                           p=[
                              kl_b[p1[i]][j]/kl_b_sum[j],
                              kl_b[p2[i]][j]/kl_b_sum[j],
                             ]
                          )
                cur[i,choices] =  cur[p1][i,choices]
                cur[i,~choices] = cur[p2][i,~choices]
        else:
            choices = self.rs.choice([True,False],size=(cur.shape))
            cur[choices] =  cur[p1][choices]
            cur[~choices] = cur[p2][~choices]
        # Re-evaluate likelihood
        L_cur = L_fn(cur, **kwargs)

        # Hold best guesses over through breeding
        drop_index = np.argsort(L_cur)[:n_carry]
        cur[drop_index] = carry
        L_cur[drop_index] = L_carry
        # Identify new best guesses
        carry_index = np.argsort(L_cur)[-n_carry:]
        carry = cur[carry_index]
        L_carry = L_cur[carry_index]

        ## Generate new steps ##
        # Initialize new samples to keep
        new = np.zeros(sample_shape)
        # loop through each random walker
        for i, item in enumerate(cur):
            # Find a valid future step
            new[i] = multivariate_normal.rvs(item,sig,random_state=self.rs)
        keep = self.satisfies_constraints(new)
        new[~keep] = cur[~keep]

        # Determine the likelihood of the new guess
        L_new = L_fn(new, **kwargs)

        # Determine alpha
        alpha = L_new/L_cur

        # Decide if to jump
        jumpseed = (self.rs.uniform(size=nwalk) > (1 - alpha)).astype(bool)
        jumpseed[L_cur==0] = True

        # Jump
        new[~jumpseed] = cur[~jumpseed]
        L_new[~jumpseed] = L_cur[~jumpseed]

        # Hold best guesses
        drop_index = np.argsort(L_new)[:n_carry]
        new[drop_index] = carry
        L_new[drop_index] = L_carry

        return new.copy(), L_new.copy()

    def RW_annealing_step(self, *args, it=0, **kwargs):
        # Load useful hyperparameters
        ntau = kwargs["ntau"]
        nstep = kwargs["nstep"]
        jump_scale = kwargs["jump_scale"]
        # Estimate step scale
        step_scale = jump_scale*np.exp(-(ntau*it)/nstep)
        # do a step
        return self.RW_normal_step(*args, scale=step_scale, it=0, **kwargs)

    #### Random Walk Algorithms ####

    def fit_random_walk(self, verbose=False, **kwargs):
        '''\
        Begin using a random walk to find the MLE value for our model
        '''
        ## Imports ##
        # Public
        import time
        import numpy as np
        from numpy.linalg import eigvals, eigvalsh
        # Check method
        method = kwargs["fit_method"]
        if method == "uniform":
            rw_method = self.RW_uniform_step
        elif method == "normal":
            rw_method = self.RW_normal_step
        elif method == "annealing":
            rw_method = self.RW_annealing_step
        elif method == "genetic":
            rw_method = self.RW_genetic_step
        else:
            raise RuntimeError("Invalid method")

        if verbose:
            print("Initiating random walk!") 
        # Draw out hyperparameters for ease of use
        nwalk = kwargs["nwalk"]
        nstep = kwargs["nstep"]

        ## Generate initial placements for MCMC walkers in parameter space ##
        #cur = self.sample_uniform(nwalk)
        if verbose:
            print("Initializing placements for random walkers")
        t_start = time.time()
        cur, fn = self.walker_init(**kwargs)
        t_end = time.time()
        if verbose:
            print("Random Walkers Initialized! (%f seconds!)"%(
                t_end - t_start))

        ## Initialize MLE structure ##
        #MLE_params = np.zeros((nwalk,nstep,len(self._parameters)))
        #MLE_likelihood = np.zeros((nwalk,nstep))

        # Initialize the best fit
        L_cur = fn(cur, **kwargs)
        index = np.argmax(L_cur)
        best_guess = cur[index].copy()
        L_best = L_cur[index]
        L_init = L_best

        # Do the fit
        t_start = time.time()
        for i in range(nstep):
            cur, L_cur = rw_method(fn, cur.copy(), it=i,**kwargs)
            # Testing
            if np.max(L_cur) > L_best:
                j = np.argmax(L_cur)
                if self.satisfies_constraints(cur[j,:]):
                    best_guess = cur[j,:].copy()
                    L_best = L_cur[j]
                    limits = self.read_limits()
                    if verbose:
                        print("\nBest fit update:\tstep %d\tkl: %f\tL: %f"%(
                                i, np.power(L_best, -1.), L_best))
                        print(best_guess)
                        print("Valid fraction: %f"%(np.sum(L_cur > 0.)/L_cur.size))
                        print("Relative variance: ")
                        print(np.std(cur,axis=0)/(limits[:,1] - limits[:,0]))
                        #print("kl_div by component:")
                        #print(self.mesh_kl(
                        #                   kwargs["mesh"],
                        #                   best_guess,
                        #                   mode="component",
                        #                   **kwargs
                        #                  ))
                        #print("eig: ")
                        #cov = cov_of_params(best_guess)
                        #print(eigvalsh(cov))
                        #if "ntau" in kwargs:
                        #    step_scale = kwargs["jump_scale"] * \
                        #            np.exp(-(kwargs["ntau"]*i)/nstep)
                        #    sig = step_scale*(limits[:,1] - limits[:,0])
                        #    print(sig/(limits[:,1] - limits[:,0]))
                        #    #print("\twalker var: %f\tstep scale: %f"%(
                        #    #        np.var(L_cur), step_scale))

        t_end = time.time()
        if verbose:
            print("Random walk time: %f seconds"%(t_end - t_start))

        if (L_best == L_init) and verbose:
            print("Did not make any progress from existing fit")
            return

        # Assign the best guess!
        if verbose:
            print("\nAssigning guess")
            print(best_guess)
            cov = cov_of_params(best_guess)
            print("eig: ")
            print(eigvalsh(cov))
        self.assign_guess(best_guess)#, force=True)

        # Find the pdf of the best guess
        #P_cur = self.likelihood(Y, return_prod=False, log_scale=False)

        return

